#%%[markdown]
# ### 聚类分析
# > 通过聚类方法，分析ProtRe-Clane，展现方法对基方法的优化和改进

# > 具体分析方法为，将每种基方法生成的sequence-superfamily打分，看成每种序列的特征向量，基于该特征向量，在指定聚类个数的条件下，使用多种聚类算法，分析聚类后的结果是否匹配序列对应的超家族

# > 评估方法为auc，因为聚类算法可以输出每个sequence所属的label, 把sequence与sequence的label进行比较得到binary prediction, 将该结果与序列真实结果相比较，即序列是否属于同一个superfamily，计算auc

#%%[markdown]

# ### 聚类的统一config
#%% [markdown]
# ### 归一化函数
# > 方便特征用于聚类


def norms(tensor, ords, axis):
    norm_list = [
        tf.norm(tensor,ord=x,axis=axis,keepdims=True)\
            for x in ords
    ]
    norms = tf.concat(norm_list,axis=axis)
    return norms

def normalization(con, axis):
    cnorms = norms(con,[1,2,np.inf],axis)
    cnorm = tf.math.reduce_mean(cnorms,axis,keepdims=True)
    con /= cnorm
    return con

def clu_config_template(specific_func):
    def clu_config_func(feature_matrix_path,n_jobs=10,**kArgs):
        feature_matrix_path = feature_matrix_path
        cross_validation_path = "dataset/benchmark/cross/fold_num2sequence_name.csv"
        key2value_path = "dataset/scop_scope/csv/sequence_superfamily.csv"
        specific_path = "feature/hhblits/acc_hmm/sequence_name2row_num.csv"
        ispecific_path = "feature/hhblits/acc_hmm/ind_sequence_name2row_num.csv"
        params = dict(
            n_jobs=n_jobs
        )
        params.update(specific_func(**kArgs))

        config = dict(
            feature_matrix_path = feature_matrix_path,
            cross_validation_path = cross_validation_path,
            key2value_path = key2value_path,
            specific_path = specific_path,
            ispecific_path = ispecific_path,
            params = params,
            decfunc = None,
            prefunc= None
        )
        return config
    return clu_config_func

def pairwise_metric_template(metric_func):
    def pairwise_func(classes, labels):
        bicla = label_binarize(classes,classes)
        bilbel = label_binarize(labels,labels)
        return metric_func(bicla,bilbel)
    return pairwise_func

#%%[markdown]
# ### 降维
def decomposition(feamat):
    def call(*args):
        dcalgo, params = args
        model = dcalgo(**params)
        dcmat = model.fit_transform(feamat)
        return dcmat
    return call

#%%

# 针对这个函数，可以使用map，因为不同聚类算法针对的feature matrix都是相同的

def clustering(feamat):
    def call(*args):
        clualgo, params = args
        model = clualgo(**params)
        labels = model.fit_predict(feamat)
        return labels
    return call

#%% [markdown]
# ### LTR预测结果转为聚类结果
class LTR(object):
    def __init__(self,**params):
        self.params = params
    
    def fit_predict(self,feamat):
        labels = tf.argsort(feamat,axis=1,direction='DESCENDING')
        return tf.transpose(labels)[0].numpy()

# %% [markdown]
# ```python
# sklearn.cluster.AgglomerativeClustering(n_clusters=2, *, affinity='euclidean', memory=None, connectivity=None, compute_full_tree='auto', linkage='ward', distance_threshold=None)
# sklearn.cluster.DBSCAN(eps=0.5, *, min_samples=5, metric='euclidean', metric_params=None, algorithm='auto', leaf_size=30, p=None, n_jobs=None)
# sklearn.cluster.OPTICS(*, min_samples=5, max_eps=inf, metric='minkowski', p=2, metric_params=None, cluster_method='xi', eps=None, xi=0.05, predecessor_correction=True, min_cluster_size=None, algorithm='auto', leaf_size=30, n_jobs=None)
# sklearn.cluster.Birch(*, threshold=0.5, branching_factor=50, n_clusters=3, compute_labels=True, copy=True)
# ```

# > 选用DBSCAN, OPTICS是因为，这两种都是基于密度的聚类算法，且不用指定聚类个数，适用于较大规模数据的聚类

#%% [markdown]
# ### 运行单一聚类算法
# > 从0.5到0.01,auc从0.52提升到了0.6977

# > 0.01时聚类个数为166

# > 0.001的时候auc降到了0.4557,聚类个数降到了71

# > 0.005, auc: 0.6234, 聚类个数115

# >
# | eps | auc | 聚类个数|
# |:----|:------|:-----|
# |0.5 | 0.52 | 未测试|
# |0.01| 0.6977 | 166|
# |0.001|0.4557 | 71|
# |0.005 | 0.6234|115|
# |0.007 | 0.6510|152|
# |0.05 | 0.7631 | 134|
# |0.02 | 0.7486 | 177|
# |0.07 | 0.7526 | 123|
# |0.06 | 0.75524|126|
# |0.055 | 0.7637|131|
# |0.053 | 0.7635|131|

# > eps=0.055
# >
# | min_samples | auc | 聚类个数|
# |:----|:------|:------|
# |100|0.68857|3
# |3 | 0.7627|227
# |10|0.7617|56
# |4 | 0.7624|168
# |6 | 0.7628|106
# |5 | 0.7637|131


# > eps = 0.055, min_samples = 5, metric = l2
# 没啥变化
# 将outliters也作为一个聚类
def outliters2clusters(labels):
    max_label = np.max(labels)
    # 找出outliters
    indices = np.where(labels == -1)
    outliter_size = indices[0].shape[0]
    outliters_labels = np.arange(max_label+1,max_label+outliter_size+1,1)
    labels[indices] = outliters_labels
    return labels
    

def single_clu(clua, cv_fold, **kArgs):
    i = cv_fold
    fn_col = str(i)
    decfunc=kArgs["decfunc"]
    prefunc=kArgs["prefunc"]
    params=kArgs["params"]
    feature_matrix_path = kArgs["feature_matrix_path"]
    key2value_path = kArgs["key2value_path"]
    cross_validation_path = kArgs["cross_validation_path"]
    specific_path = kArgs["specific_path"]
    ispecific_path = kArgs["ispecific_path"]

    cv_df = pd.read_csv(cross_validation_path)
    kv_df = pd.read_csv(key2value_path)
    spec_df = pd.read_csv(specific_path,index_col=0)
    ispec_df = pd.read_csv(ispecific_path,index_col=0)

    template_df = spec_df.reset_index().rename(columns={"index":"template"})
    itemplate_df = ispec_df.reset_index().rename(columns={"index":"template"})

    im_df = itemplate_df.join(kv_df.set_index("template"),on="template")
    im_df = im_df.sort_values(by="row num")
    classes = im_df.loc[:,"label"].to_numpy(int)

    train_df = template_df.join(kv_df.set_index("template"),on="template")
    t_i_df = train_df.loc[~train_df.loc[:,"template"].isin(cv_df.loc[:,fn_col])]
    # 重新编号
    sfdf = pd.DataFrame(list(t_i_df.groupby("label").groups.keys()),columns=["label"])
    sfdf = sfdf.reset_index().rename(columns={"index":"new_label"})


    feamat = np.load(feature_matrix_path)
    # 去除负权值边
    # 使用norm反而不好\
    # feamat[feamat < 0] = 0
    if np.any(feamat < 0):
        min_fea_score = feamat.min()
        feamat += np.abs(min_fea_score)
        absmin_score = np.abs(feamat[feamat > 0]).min()
        # 加一个最小的值
        # 保证不为0
        epsilon = absmin_score * 10**(-round(absmin_score,0))
        feamat += epsilon
        # feamat[feamat < 0] = 0
        norm = np.linalg.norm(feamat,np.inf,axis=1,keepdims=True)
        feamat /= norm
    # 加上一个最小量保证没有0

    # feamat = np.exp(feamat)

    # 降维
    if decfunc is not None:
        feamat = decfunc(feamat)
    
    if prefunc is not None:
        feamat = prefunc(feamat)
        feamat = feamat.numpy()

    labels = clustering(feamat)(clua,params)

    # # 对应为原来的标签
    # lbel_df = pd.DataFrame(labels,columns=["new_label"])
    # lbel_df = lbel_df.join(sfdf.set_index("new_label"),on="new_label")
    # labels = lbel_df.loc[:,"label"].to_numpy()

    labels = outliters2clusters(labels)


    return classes, labels

#%% [markdown]

# >
# | n_components | auc | 聚类个数
# |:----|:-----|:---|
# | 0.9 | 0.6857|56|
# | 0.95 | 0.7120|78|
# | 0.99 | 0.7441|119|
# | 0.999 | 0.7584|129|
# | 0.9999 | 0.7626|131|

def single_dec(feamat):
    deca = PCA

    params = dict(n_components = 0.9999)
    return decomposition(feamat)(deca,params)


#%% [markdown]
# ### 通过auc评估聚类结果

# > classes表示序列真正的superfamily, labels表示聚类的结果

# >> 注意不能直接比较，因为，classes和labels里面的内容是对不上的
# 同时要注意分析outliters
# 把单个outliters重新给个新标签
# @pairwise_metric_template
def clu_metric_auc(classes, labels):
    y_true = label_binarize(classes,classes)
    y_pred = label_binarize(labels,labels)
    score = roc_auc_score(y_true,y_pred)
    return score

# @pairwise_metric_template
def clu_metric_acc(classes, labels):
    y_true = label_binarize(classes,classes)
    y_pred = label_binarize(labels,labels)
    score = accuracy_score(y_true,y_pred)
    return score



# %% [markdown]
# 不使用降维效果会更好
# 
# |Mehtod | eps |
# |:------------|:------|
# | LinearSVM(HMM) | 0.055| 0.7656|
# | LinearSVM(PSSM) | 0.055 | 0.7691|
# | DFG-S2C(HHblits) | 0.055 | 0.9297|
# | ProtRe-Clane | 0.7| 0.8737|
# | ProtRe-Clane | 0.9| 0.8803|
# | ProtRe-Clane | 1.1 |0.8844|
# | ProtRe-Clane | 1.5 |0.8909|
# | ProtRe-Clane | 1.7 |0.8879|
# | ProtRe-Clane | 1.65|0.8875|
# | ProtRe-Clane | 1.59|0.8877|
# | ProtRe-Clane | 1.53|0.8913|
#
# 
# >使用归一化后
# 
# |Mehtod | eps |
# |:------------|:------|
# | LinearSVM(HMM) | 0.0055| 0.8170|
# | DFG-S2C(HHblits) | 0.055 | 0.9297|
#
# > 将LTR结果转为聚类结果
#
# |Mehtod | auc |
# |:------------|:------|
# | LinearSVM(HMM)|0.8253|
# | LinearSVM(PSSM)|0.8137|
# | ProtRe-Clane | 0.9788|


@clu_config_template
def dbscan_config(**kArgs):
    eps = kArgs["eps"]
    params = dict(eps=eps,min_samples=5)
    return params

#%%
# 这里就用cluster_method="dbscan"
# 效果和直接用DBSCAN差不多
# 但速度非常慢
@clu_config_template
def optics_config(**kArgs):
    params = dict(eps=0.055,min_samples=5,cluster_method="dbscan")
    return params
#%%
def call_clus():
    n = 5 # 表示5折交叉验证
    feature_mat_dirs = [
        "feature/multiclassification/training/hhblits/representation",
        "feature/multiclassification/training/psiblastexb/representation",
        "feature/S2C/training_cluster/hhblitsexb",
        "feature/S2C/training_cluster/hhsearchexb",
        "feature/S2C/training_cluster/psiblastexb",
        "feature/S2C/training_cluster/deltablast",
        "feature/LTR/training/representation"  
    ]

    n_jobs = 20
    eps_list = [0.9,0.9,0.1,0.1,0.06,0.06,0.8625]
    params_list = list(
        map(lambda e:dict(eps=e),eps_list)
    )
    clua = DBSCAN

    def cv_clu(cv_fold):
        feature_mat_paths = list(map(lambda d:os.path.join(d,"ind0{}.npy".format(cv_fold)),feature_mat_dirs))
        def iter_clu(round):
            feature_matrix_path = feature_mat_paths[round]
            params = params_list[round]
            return single_clu(
                clua,cv_fold,**dbscan_config(
                    feature_matrix_path,n_jobs,**params
                )
            )

        
        return list(map(iter_clu,range(len(eps_list))))
    
    return list(map(cv_clu,range(n)))
        

def call_single_clu():
    s = time.time()
    # feature_matrix_path = "feature/multiclassification/training/hhblits/representation/ind00.npy"
    # feature_matrix_path = "feature/multiclassification/training/psiblastexb/representation/ind00.npy"
    # feature_matrix_path = "feature/S2C/training_cluster/hhblitsexb/ind00.npy"
    # feature_matrix_path = "feature/S2C/training_cluster/hhsearchexb/ind00.npy"
    # feature_matrix_path = "feature/S2C/training_cluster/psiblastexb/ind00.npy"
    # feature_matrix_path = "feature/S2C/training_cluster/deltablast/ind00.npy"
    feature_matrix_path = "feature/LTR/training/representation/ind00.npy"

    n_jobs = 20
    cv_fold = 0
    
    # params = dict(eps=0.02) # LinearSVM(HMM)
    # params = dict(eps=0.03) # LinearSVM(PSSM)
    # params = dict(eps=0.005) # DFG-S2C(HHblits), DFG-S2C(HHsearch)
    # params = dict(eps=0.03) # DFG-S2C(PSI-BLAST), DFG-S2C(DELTA-BLAST)
    # params = dict(eps=1.53) # ProtRe-Clane

    # 特征处理且经过归一化
    # params = dict(eps=0.9) # LinearSVM(HMM),  LinearSVM(PSSM)
    # params = dict(eps=0.1) # DFG-S2C(HHblits), DFG-S2C(HHsearch)
    # params = dict(eps=0.06) # DFG-S2C(PSI-BLAST),DFG-S2C(DELTA-BLAST)
    params = dict(eps=0.8625) # ProtRe-Clane
    
    classes, labels = single_clu(DBSCAN,cv_fold,**dbscan_config(feature_matrix_path,n_jobs=n_jobs,**params))
    # classes, labels = single_clu(LTR,**dbscan_config(feature_matrix_path,n_jobs=n_jobs,**params))
    e = time.time()
    # print(e-s)
    return classes, labels

def get_cluster_num(labels):
    g = groupby(sorted(labels))
    def fetch(*args):
        x,_ = args
        return x

    L = list(map(lambda p: fetch(*p), g))
    return L

def cons_mfunc(metric):
    def call(y_true,y_pred):
        return metric(y_true,y_pred)
    return call
    
def metric_vectorize(func):
    def call(*args):
        y_true_mat, y_pred_mat = args
        def iter_metric(row):
            y_true = y_true_mat[row]
            y_pred = y_pred_mat[row]
            return func(y_true,y_pred)
        return list(map(iter_metric,range(y_true_mat.shape[0])))
    return call
# 对所有聚类结果进行评估
# tensor大小为k * 2 * n
# k表示k种基方法
# 2表示y_true, y_pred
# n表示n_samples
def cv_metrics(tensor):
    metrics = [
        normalized_mutual_info_score,adjusted_mutual_info_score
    ]
    
    tnsr = np.transpose(tensor,[1,0,2])
    
    mfunc_list = list(map(cons_mfunc,metrics))
    vfunc_list = list(map(metric_vectorize,mfunc_list))
    def call_vfunc(m_i):
        vfunc = vfunc_list[m_i]
        return vfunc(*tnsr)
    smat_list = list(
        map(call_vfunc,range(len(metrics)))
    )
    return smat_list

def cv_metric(tensor):
    metric = homogeneity_completeness_v_measure
    tnsr = np.transpose(tensor,[1,0,2])
    mfunc = cons_mfunc(metric)
    vfunc = metric_vectorize(mfunc)
    return vfunc(*tnsr)

# tensor_lsit包含n_fold个tensor
# n_fold表示交叉验证划分的个数
def clu_metrics(tensor_list):
    return list(map(cv_metrics,tensor_list))


def clu_metric(tensor_list):
    return list(map(cv_metric,tensor_list))

def use_single_call_clu():
    classes, labels = call_single_clu()
    print(normalized_mutual_info_score(classes,labels))
    print(adjusted_mutual_info_score(classes,labels))
    print(homogeneity_completeness_v_measure(classes, labels))
    print(len(get_cluster_num(labels)))

def use_call_clus():
    clable_list = call_clus()
    tensor_list = list(map(np.array,clable_list))
    smat_list = clu_metrics(tensor_list)
    smat_ary = np.array(smat_list)
    return smat_ary

def save_clu_result(tensor_list):
    n = 5
    clu_dirs = [
        "evaluation/cluster/multiclassification/hhblits",
        "evaluation/cluster/multiclassification/psiblastexb",
        "evaluation/cluster/S2C/hhblitsexb",
        "evaluation/cluster/S2C/hhsearchexb",
        "evaluation/cluster/S2C/psiblastexb",
        "evaluation/cluster/S2C/deltablast",
        "evaluation/cluster/LTR"
    ]

    def cv_paths(cv_fold):
        clu_paths = list(map(lambda d: os.path.join(d,"ind0{}.npy".format(cv_fold)),clu_dirs))
        return clu_paths
    
    clu_paths_list = list(map(cv_paths,range(n)))

    def save_cv(cv_fold):
        clu_paths = clu_paths_list[cv_fold]
        tensor = tensor_list[cv_fold]
        def save_method(m):
            tm = tensor[m]
            clu_path = clu_paths[m]
            np.save(clu_path,tm)

        return list(map(save_method,range(len(clu_dirs))))
    
    list(map(save_cv,range(n)))

def read_clu_result():
    n = 5
    clu_dirs = [
        "evaluation/cluster/multiclassification/hhblits",
        "evaluation/cluster/multiclassification/psiblastexb",
        "evaluation/cluster/S2C/hhblitsexb",
        "evaluation/cluster/S2C/hhsearchexb",
        "evaluation/cluster/S2C/psiblastexb",
        "evaluation/cluster/S2C/deltablast",
        "evaluation/cluster/LTR"
    ]

    def cv_paths(cv_fold):
        clu_paths = list(map(lambda d: os.path.join(d,"ind0{}.npy".format(cv_fold)),clu_dirs))
        return clu_paths
    
    clu_paths_list = list(map(cv_paths,range(n)))

    def read_cv(cv_fold):
        clu_paths = clu_paths_list[cv_fold]
        def read_method(m):
            clu_path = clu_paths[m]
            return np.load(clu_path)

        return list(map(read_method,range(len(clu_dirs))))
    
    return list(map(read_cv,range(n)))

def save_clu_metrics(smat_list):
    n = 5
    clu_dir = "evaluation/clu_metric"
    def cv_path(cv_fold):
        clu_path = os.path.join(clu_dir,"ind0{}_clu_metrics.txt".format(cv_fold))
        return clu_path
    
    clu_paths = list(map(cv_path,range(n)))

    def save_cv(cv_fold):
        clu_path = clu_paths[cv_fold]
        smat = smat_list[cv_fold]
        np.savetxt(clu_path,smat,"%.16f")
    
    list(map(save_cv,range(n)))


#%%
if __name__ == "__main__":
    from sklearn.cluster import DBSCAN, OPTICS
    from sklearn.preprocessing import label_binarize
    from sklearn.metrics import roc_auc_score, accuracy_score
    from sklearn.metrics import  adjusted_mutual_info_score,normalized_mutual_info_score
    from sklearn.metrics import homogeneity_completeness_v_measure
    from sklearn.decomposition import PCA
    import numpy as np
    import pandas as pd
    import tensorflow as tf
    import time
    from itertools import groupby
    import os

    # use_single_call_clu()

    # 多种Metric
    tensor_list = call_clus()
    tensor_list = list(map(np.array,tensor_list))
    save_clu_result(tensor_list)
    smat_list = clu_metrics(tensor_list)
    smat_ary = np.array(smat_list)
    print(smat_ary.mean(0))
    save_clu_metrics(smat_list)

    # 读取结果，使用单个的metric
    tensor_list = read_clu_result()
    smat_list = clu_metric(tensor_list)
    smat_ary = np.array(smat_list)
    print(smat_ary.mean(0))
    
#%%

# %%

# %%
