# 通过igraph绘制框架的核心思想

#%%
# 抽样，抽取100个superfamily

def sample_superfamily(n_sf, sequence_identity, kv_df, diffdf):
    ser = diffdf.loc[diffdf.loc[:,sequence_identity],"template"]
    dkv_df = kv_df.loc[kv_df.loc[:,"template"].isin(ser)]
    g = dkv_df.groupby("label")
    superfams = list(g.groups.keys())
    sfser = pd.Series(superfams)
    samplesf_ser = sfser.sample(n_sf)
    sf_df = g.apply(lambda x:x.loc[x.index[0]])
    samplesf_df = sf_df.loc[sf_df.loc[:,"label"].isin(samplesf_ser)]
    samplesf_df.reset_index(drop=True,inplace=True)
    samplesf_df.drop(columns=["template","sf"],inplace=True)
    samplesf_df.loc[:,"new_label"] = samplesf_df.index
    samplekv_df = dkv_df.loc[dkv_df.loc[:,"label"].isin(samplesf_ser)]
    samplekv_df = samplekv_df.join(samplesf_df.set_index("label"),on="label")
    samplekv_df.reset_index(drop=True,inplace=True)
    return samplekv_df

# 把sf同样看作node
# 建立统一的编号
# label, sf为NAN表示对应的node为superfamily，只是在绘图中也被视为顶点
def renumer(samplekv_df):
    sfser = pd.Series(list(samplekv_df.groupby("sf").groups.keys()))
    template_ser = samplekv_df.sort_values(by="template").loc[:,"template"]
    allser = pd.concat([sfser,template_ser],axis=0)
    alldf = pd.DataFrame(allser,columns=["name"],index=range(allser.shape[0]))
    alldf.loc[:, "row num"] = alldf.index
    alldf = alldf.join(samplekv_df.set_index("template"),on="name")
    return alldf

# 建立matrix
# n_names * n_names大小
# 包含sequece与sequence之间，sequence与superfamily之间，superfamily与superfamily之间
# 权值分别为1,0.1,1
# 但要注意分别属于scop和scope之间的序列，权值同样为0
def construct_mat(alpha, beta, alldf, samplekv_df, scop_df, scope_df):
    n_names = alldf.shape[0]
    indices = alldf.loc[pd.isna(alldf.loc[:,"label"]),"row num"]
    n_sub = indices.shape[0]
    # superfamily与superfamily之间的权值
    w0 = np.ones((n_sub,n_sub))
    # sequence与sequence之间的权值
    w3 = np.ones((n_names-n_sub,n_names-n_sub))
    # sequence与superfamily之间的权值
    w2 = alpha * np.ones((n_names-n_sub, n_sub))
    # superfamily与sequence之间的权值
    w1 = np.zeros((n_sub,n_names-n_sub))

    wmat = np.concatenate(
        [
            np.concatenate(
                [w0,w1],axis=1
            ),
            np.concatenate(
                [w2,w3],axis=1
            )
        ],
        axis=0
    )

    # 属于不同数据集sequence的权值应该为比0.1更小的值
    indices0 = alldf.loc[alldf.loc[:,"name"].isin(scop_df.loc[:,"template"]),"row num"]
    indices1 = alldf.loc[alldf.loc[:,"name"].isin(scope_df.loc[:,"template"]),"row num"]
    l0 = indices0.to_numpy(int)
    l1 = indices1.to_numpy(int)
    wmat[np.ix_(l0,l1)] = alpha - beta
    wmat[np.ix_(l1,l0)] = alpha - beta

    # sequence与对应的superfamily之间应该有更大的权重
    # 但也不能太大设置为1+alpha
    seq_indices = alldf.loc[~pd.isna(alldf.loc[:,"new_label"]),"row num"].to_numpy(int)
    sf_indices = alldf.join(
        samplekv_df.set_index("template"),on="name",rsuffix="_overlap"
    ).dropna(axis=0).loc[:,"new_label_overlap"].to_numpy(int)
    indices = np.stack([seq_indices,sf_indices],axis=0).T
    updates = (1+alpha)*tf.ones(indices.shape[0])
    wmat = tf.tensor_scatter_nd_update(wmat,indices,updates).numpy()

    # 去掉自环
    wmat[range(n_names),range(n_names)] = 0


    return wmat,l0,l1

def save_mat(mat, save_path):
    np.save(save_path,mat)

def save_mat_txt(mat,save_path):
    np.savetxt(save_path,mat,"%f")

# 将矩阵转为边表格和node表格
# dataframe格式
# columns为Source	Target	Type	Id	Label	timeset	Weight
def mat2csv(wmat, alldf):
    indices = np.stack(np.where(wmat > 0),axis=0).astype(int).T
    inddf = pd.DataFrame(indices,columns=["Source","Target"])
    inddf.loc[:,["Type","Id","Label","timeset"]] = None
    inddf.loc[:,"Weight"] = tf.gather_nd(wmat,indices).numpy()
    noddf = alldf.loc[:,["row num","name"]].rename(columns={"row num":"Id","name":"Label"})
    noddf.loc[:,"timeset"] = None
    return noddf, inddf

def save_csv(df, save_path):
    df.to_csv(save_path,index=False)

def save_fgraph_data():
    the_num_of_sf = 10
    the_sequence_identity = "0.95"
    the_key2value_path = "dataset/scop_scope/csv/sequence_superfamily.csv"
    the_diff_path = "dataset/scop_scope/csv/sequence_identity.csv"
    alpha = 0.5
    beta = 0.1
    the_scop_path = "dataset/SCOP/csv/sequence_superfamily.csv"
    the_scope_path = "dataset/SCOPe/csv/sequence_superfamily.csv"
    the_save_path = "visualization/data/framework/weighted_matrix.npy"
    the_savetxt_path = "visualization/data/framework/weighted_matrix.txt"
    the_nodes_path = "visualization/data/framework/wmat_nodes.csv"
    the_edges_path = "visualization/data/framework/wmat_edges.csv"
    the_scop_nds_path = "visualization/data/framework/wmat_scop_nodes.csv"
    the_scope_nds_path = "visualization/data/framework/wmat_scope_nodes.csv"

    the_kv_df = pd.read_csv(the_key2value_path)
    the_diffdf = pd.read_csv(the_diff_path)
    the_samplekv_df = sample_superfamily(the_num_of_sf,the_sequence_identity,the_kv_df,the_diffdf)
    the_alldf = renumer(the_samplekv_df)
    the_scop_df = pd.read_csv(the_scop_path)
    the_scope_df = pd.read_csv(the_scope_path)
    the_wmat, the_l0, the_l1 = construct_mat(alpha,beta,the_alldf,the_samplekv_df,the_scop_df,the_scope_df)
    # save_mat(the_wmat,the_save_path)
    save_mat_txt(the_wmat,the_savetxt_path)
    noddf, inddf = mat2csv(the_wmat,the_alldf)
    save_csv(noddf,the_nodes_path)
    save_csv(inddf,the_edges_path)

    the_scop_nd_df = the_alldf.loc[the_alldf.loc[:,"row num"].isin(the_l0)]
    the_scope_nd_df = the_alldf.loc[the_alldf.loc[:,"row num"].isin(the_l1)]

    save_csv(the_scop_nd_df,the_scop_nds_path)
    save_csv(the_scope_nd_df,the_scope_nds_path)


# 分析数据集中远同源与非远同源的比例
def fasta2dict(level, fas_path):
    from Bio.SeqIO import parse
    
    if level == "fold":
        n_pos = 1
    elif level == "sf":
        n_pos = 2
    elif level == "family":
        n_pos = 3
    else:
        assert level in ["fold", "sf", "family"]
    
    # 把description转为protein对应的type，比如superfamily
    def des2type(des):
        import re

        searcher = re.compile(r"\w+(\.\w+){%d}" % n_pos)
        type = searcher.search(des)
        return type.group()
    
    def recoder2sfseq(r):
        return des2type(r.description), r.name

    with open(fas_path,"r") as h:
        sfseq_list = list(map(recoder2sfseq,parse(h,"fasta")))
    
    sfdict = {}

    def cons_sfdict(*args):
        sf, seq = args
        if sfdict.get(sf) is None:
            sfdict[sf] = []
        sfdict[sf].append(seq)
    
    list(map(lambda p: cons_sfdict(*p), sfseq_list))
    return sfdict

# 为了能展现数据集不同superfamily的分布特征
# 需要转为graph
# 不同superfamily之间本来是没有联系的，而且也不好展现大小
# 因此，这里订一个规则，就是让包含序列少的指向包含序列多的superfamily
# 这样序列越多的superfamily，degree就会越高，这样就可通过degree调整点的大小
def dict2graph(sfdict, kv_df):
    def get_length(*args):
        sf, seq_list = args
        return sf, len(seq_list)
    sz_df = pd.DataFrame(list(map(lambda p: get_length(*p),sfdict.items())),columns=["sf","size"])
    vdf = kv_df.groupby("sf").apply(lambda d:d.loc[d.index[0]])
    sz_df = sz_df.join(vdf.loc[:,["sf","label"]].set_index("sf"),on="sf")

    n_sf = sz_df.shape[0]
    # 根据label重新编新的label，作为row num
    sz_df.sort_values(by="label",inplace=True)
    sz_df.reset_index(drop=True,inplace=True)
    sz_df.loc[:,"new_label"] = sz_df.index

    # 根据size分组，这样方便建立graph
    g = sz_df.groupby("size")["new_label"]
    szmat = np.zeros((n_sf,n_sf))

    nlary_list = list(map(lambda p:p[1].to_numpy(dtype=int),g))

    # 因为nlary_list是根据size从小到大排序过的
    # 因此只要顺序处理即可
    def cons_mat(i):
        # 获取size较小的superfamily对应的row num
        indices0 = nlary_list[i]
        # 获取size较大的superfamily对应的row num
        indices1 = np.concatenate(nlary_list[i:],axis=0)
        szmat[np.ix_(indices0,indices1)] = 1
    
    list(map(cons_mat,range(len(nlary_list))))

    # 为了后续方便处理数据
    sz_df.rename(columns={"new_label":"row num","sf":"name"},inplace=True)

    return szmat, sz_df.loc[:,["name","row num"]]

def save_sfgraph_data():
    level = "sf"
    fas_path = "dataset/benchmark/all/independent"
    key2value_path = "dataset/scop_scope/csv/sequence_superfamily.csv"
    superfamily_nodes_path = "visualization/data/superfamily/superfamily_nodes.csv"
    superfamily_edges_path = "visualization/data/superfamily/superfamily_edges.csv"

    kv_df = pd.read_csv(key2value_path)
    sfdict = fasta2dict(level,fas_path)

    szmat, sz_df = dict2graph(sfdict,kv_df)
    # print(sz_df)
    noddf,inddf = mat2csv(szmat,sz_df)
    noddf.to_csv(superfamily_nodes_path,index=False)
    inddf.to_csv(superfamily_edges_path,index=False)

#%%[markdown]
# ### 分析基方法中的false positive
# > 目的是为了展现模型的纠错能力

def get_data_shape(fpath):
    return np.load(fpath).shape[0]

def split_func(p,x):
    L, start, bound, result, iter = p
    if x < bound:
        L.append(x-start)
        return L,start, bound, result, iter
    else:
        start = bound
        bound = next(iter)
        result.append(L)
        return [x-start], start, bound, result,iter

# pred是评估结果
# 有可能3列也有可能两列
# 但必然包含roc1, roc50
# 3列的结果，多包含了一列表示auc
def get_fp_boolean_mask(evaluation_path):

    emat = np.loadtxt(evaluation_path,delimiter=" ")
    if emat.shape[1] == 3:
        emat = emat[:,:2]
    
    # # 先找出为0的indices
    # bmask = (emat == 0)

    return emat[:,0]


def fp_data(**kArgs):
    evaluation_paths = kArgs["evaluation_paths"]

    indices_list = list(map(get_fp_boolean_mask,evaluation_paths))

    # k * n_selected
    # k = 6
    # 6表示6种基方法
    # n_selected表示选择的false positive个数
    indices_mat = np.stack(indices_list,axis=0)
    # n_selected * k
    indices_mat = indices_mat.T

    # 全为false的结果
    all_fp_con = tf.zeros(indices_mat.shape)
    allfp_boolean_mask = tf.math.reduce_all(all_fp_con == indices_mat,axis=1,keepdims=True)

    # # 找到fp所在位置
    # fp_boolean_mask = ~indices_mat
    # # 第一个为fp，其他的也都为fp
    # all_boolean_mask = tf.reshape(all_boolean_mask,(-1,1))
    # allfp_boolean_mask = (all_boolean_mask == fp_boolean_mask[:,0].reshape(-1,1))

    return allfp_boolean_mask.numpy()

def fp_sequence(**kArgs):
    save_fp_path = kArgs["save_fp_path"]
    bounds = kArgs["bounds"]
    specific_path = kArgs["specific_path"]
    allfp_imat = np.loadtxt(save_fp_path,dtype=int)
    spec_df = pd.read_csv(specific_path,index_col=0)
    indices = allfp_imat[:,0]

    biter = accumulate(bounds,operator.add)

    bound = next(biter)
    plist = reduce(split_func,indices,([],0,bound,[],biter))
    indices_slice_list = list(plist[3])
    indices_list = reduce(operator.add,indices_slice_list)

    seq_df = spec_df.loc[spec_df.loc[:,"row num"].isin(indices_list)]
    return seq_df

def call_fp_data(n):
    evaluation_paths = [
        "evaluation/roc_at/ori/multiclassification/hhblitsexb/ind_seq_rocat.txt",
        "evaluation/roc_at/ori/multiclassification/psiblastexb/ind_seq_rocat.txt",
        "evaluation/roc_at/ori/s2c/hhblitsexb/ind_seq_rocat.txt",
        "evaluation/roc_at/ori/s2c/hhsearchexb/ind_seq_rocat.txt",
        "evaluation/roc_at/ori/s2c/psiblastexb/ind_seq_rocat.txt",
        "evaluation/roc_at/ori/s2c/deltablast/ind_seq_rocat.txt"
    ]

    data_dirs = [
        "feature/LTR/training/representation",
        "feature/multiclassification/training/hhblitsexb/representation",
        "feature/multiclassification/training/psiblastexb/representation",
        "feature/S2C/training_cluster/hhblitsexb",
        "feature/S2C/training_cluster/hhsearchexb",
        "feature/S2C/training_cluster/psiblastexb",
        "feature/S2C/training_cluster/deltablast"
    ]

    fp_data_save_dirs = [
        "evaluation/false_positive/ProtRe-Clane",
        "evaluation/false_positive/LinearSVM(HMM)",
        "evaluation/false_positive/LinearSVM(PSSM)",
        "evaluation/false_positive/DFG-S2C(HHblits)",
        "evaluation/false_positive/DFG-S2C(HHsearch)",
        "evaluation/false_positive/DFG-S2C(PSI-BLAST)",
        "evaluation/false_positive/DFG-S2C(DELTA-BLAST)"
        
    ]

    prediction_path = "evaluation/roc_at/lightgbm/lambdarank/representation/ind_seq_rocat.txt"

    save_fp_path = "evaluation/false_positive/all/all_false_positive.txt"
    config = dict(evaluation_paths = evaluation_paths)
    allfp_boolean_mask = fp_data(**config)
    pred_mat = np.loadtxt(prediction_path,delimiter=" ")
    pred_tp_boolean_mask = (pred_mat[:,0]==1)
    pred_tp_boolean_mask = np.reshape(pred_tp_boolean_mask,(-1,1))

    # 选出基方法错误，但最终结果正确的
    allfp_boolean_mask = tf.math.reduce_all(tf.concat([allfp_boolean_mask,pred_tp_boolean_mask],axis=1),axis=1,keepdims=True)
    allfp_indices = np.where(allfp_boolean_mask)

    allfp_imat = np.stack(allfp_indices,axis=0).T

    np.savetxt(save_fp_path,allfp_imat,"%d")

    indices = allfp_imat[:,0]
    # 读取false positive相关的排序结果
    # 先加载所有排序结果
    fnames = list(map(lambda x :"ind0{}.npy".format(x),range(n)))
    def read_data(data_dir):
        def call(fname):
            fpath = os.path.join(data_dir,fname)
            dmat = np.load(fpath)
            return dmat
        # dmat = np.concat(
        #     list(map(call,fnames)),
        #     axis=0
        # )
        # return dmat[indices]
        mapped = list(map(call,fnames))
        return mapped

    indices_slice_list = []

    sub_data_paths = list(
        map(
            lambda x: os.path.join(data_dirs[0],"ind0{}.npy".format(x)),
            range(n)
        )
    )
    bounds = list(map(get_data_shape,sub_data_paths))
    biter = accumulate(bounds,operator.add)
    
    bound = next(biter)
    plist = reduce(split_func,indices,([],0,bound,[],biter))
    indices_slice_list = list(plist[3])
    last_one = list(plist[0])
    indices_slice_list.append(last_one)


    # fpdata_list = list(map(read_data,data_dirs))
    fp_mapped_list = list(map(read_data,data_dirs))
    # k * n_selected * n_class
    # k=7, 表示6种基方法+最终方法
    # fpdata_tensor = tf.stack(fpdata_list,axis=0)

    def save_fp_data(*args):
        data_dir, mapped = args

        def fetch_data_by_indices(*args):
            dmat, indices = args
            return dmat[indices]
        
        def iter_save(*args):
            dmat, cv_fold  = args
            save_data_path = os.path.join(data_dir,"ind0{}.npy".format(cv_fold))
            np.save(save_data_path,dmat)

        mapped = map(
            lambda p:fetch_data_by_indices(*p),
            zip(list(mapped),indices_slice_list)
        )

        mapped = map(
            lambda p: iter_save(*p),
            zip(list(mapped),range(n))
        )
        return mapped
        
    mapped_list = list(map(lambda p:save_fp_data(*p), zip(fp_data_save_dirs,fp_mapped_list)))
    list(map(list,mapped_list))

# 选取排在前select_num个的superfamily
# 建立用于绘图的数据
# fp_data_save_paths包含k个文件对应k折交叉验证实验
def cons_fpgraph_data(n, fp_data_save_paths, **kArgs):
    cv_df = kArgs["cv_df"]
    template_df = kArgs["template_df"]
    kv_df = kArgs["kv_df"]
    select_num = kArgs["select_num"]
    indices_slice_list = kArgs["indices_slice_list"]

    k_df = template_df.join(kv_df.set_index("template"),on="template").loc[:,["template","sf","label"]]
    def get_superfamily_label(cv_fold):
        fn_col = str(cv_fold)
        kv_i_df = k_df.loc[~k_df.loc[:,"template"].isin(cv_df.loc[:,fn_col])]
        label = tf.constant(list(kv_i_df.groupby("label").groups.keys()),tf.int32)
        label2new_df = pd.DataFrame(
            {"label": label,"new_label": list(range(label.shape[0]))}
        )
        return label2new_df
    
    get_sf_mapped = map(
        get_superfamily_label,
        range(n)
    )

    def load_fpdata(path):
        if os.path.exists(path):
            return np.load(path)

    get_fpdata_mapped = map(
        load_fpdata,fp_data_save_paths
    )

    def select_superfamily(*args):
        fpdata, label2new_df,indices_slice = args
        # 整体平移防止出现负值
        min_fpscore = np.min(fpdata,axis=1,keepdims=True)
        fpdata += np.abs(min_fpscore)
        # 防止出现0
        min_afpscore = np.min(np.abs(fpdata[fpdata > 0]))
        epsilon = min_afpscore * 10**(-round(min_afpscore,0))
        fpdata += epsilon

        # 归一化
        norm = np.linalg.norm(fpdata,np.inf,axis=1,keepdims=True)
        fpdata /= norm
        
        fp_argsorted = tf.argsort(fpdata,axis=1,direction="DESCENDING")
        fp_argsorted = tf.transpose(fp_argsorted)
        # 选排在前select_num的superfamily
        fp_argsorted = fp_argsorted[:select_num]
        fp_argsorted = tf.transpose(fp_argsorted)
        fp_sel = tf.reshape(fp_argsorted,(-1,))
        fp_label = label2new_df.set_index("new_label").loc[fp_sel.numpy()]
        fp_label = fp_label.to_numpy(int)
        fp_label = fp_label.reshape((-1,1))

        # 构造顶点下标组成的矩阵
        fp_rmat = np.ones((fpdata.shape[0],select_num))
        row_ary = np.array(range(fpdata.shape[0])).reshape((-1,1))
        indcies_ary = np.array(indices_slice).reshape((-1,1))
        fp_rnum = np.reshape(fp_rmat * row_ary,(-1,1))
        fp_indices = np.reshape(fp_rmat * indcies_ary,(-1,1))

        # 组合成indices
        fp_sel = tf.reshape(fp_sel,(-1,1))
        indices = np.concatenate([fp_rnum,fp_sel.numpy()],axis=1)
        indices = tf.cast(indices,dtype=tf.int32)
        # 根据indices获取数据
        fp_pred = tf.gather_nd(fpdata,indices)
        fp_pred = tf.reshape(fp_pred,(-1,1)).numpy()
        # 构建dataframe，作为后续边表格的雏形
        # row num 相当于source, label相当于target, score相当于weight
        # 这里要用fp_indices，因为，fp_rnum表示得只是在这个矩阵中的下标，这个下标并不唯一
        fpdf = pd.DataFrame(
            np.concatenate(
                [fp_indices,fp_label,fp_pred],
                axis=1
            ),
            columns=["row num","label","score"]
        )
        fpdf = fpdf.astype({"row num":int,"label":int, "score": float})
        return fpdf

    sel_fpdf_mapped = map(
        lambda p :select_superfamily(*p),
        zip(
            list(get_fpdata_mapped),
            list(get_sf_mapped),
            indices_slice_list
        )
    )

    fpdf = pd.concat(list(sel_fpdf_mapped),axis=0)
    fpdf = fpdf.reset_index(drop=True)
    # # 去重
    # # 序列对同一个superfamily的打分取最高的
    # g = fpdf.groupby(by=["row num","label"])
    # def get_max(subdf):
    #     # max_i = np.argmax(subdf.loc[:,"score"].to_numpy())
    #     max_i = subdf.loc[:,"score"].idxmax()
    #     return subdf.loc[[max_i],["score"]]
    
    # max_ser = g.apply(get_max).reset_index(drop=True).loc[:,"score"]
    # gdict = g.groups
    # gkeys_ary = np.array(list(gdict.keys()))
    # fpdf = pd.DataFrame(gkeys_ary,columns=["row num","label"])
    # fpdf.loc[:,"score"] = max_ser
    # fpdf = fpdf.reset_index()
    # fpdf = fpdf.set_index("row num")

    # 最后会多出一个level_2的列
    return fpdf

def save_fpgraph_data(n):
    cross_validation_path = "dataset/benchmark/cross/fold_num2sequence_name.csv"
    key2value_path = "dataset/scop_scope/csv/sequence_superfamily.csv"
    specific_path = "feature/hhblitsexb/acc_hmm/sequence_name2row_num.csv"
    ispecific_path = "feature/hhblitsexb/acc_hmm/ind_sequence_name2row_num.csv"
    select_num = 5

    save_fp_path = "evaluation/false_positive/all/all_false_positive.txt"
    data_dirs = [
        "feature/LTR/training/representation",
        "feature/multiclassification/training/hhblitsexb/representation",
        "feature/multiclassification/training/psiblastexb/representation",
        "feature/S2C/training_cluster/hhblitsexb",
        "feature/S2C/training_cluster/hhsearchexb",
        "feature/S2C/training_cluster/psiblastexb",
        "feature/S2C/training_cluster/deltablast"
    ]

    fp_data_save_dirs = [
        "evaluation/false_positive/ProtRe-Clane",
        "evaluation/false_positive/LinearSVM(HMM)",
        "evaluation/false_positive/LinearSVM(PSSM)",
        "evaluation/false_positive/DFG-S2C(HHblits)",
        "evaluation/false_positive/DFG-S2C(HHsearch)",
        "evaluation/false_positive/DFG-S2C(PSI-BLAST)",
        "evaluation/false_positive/DFG-S2C(DELTA-BLAST)"
    ]    

    cv_df = pd.read_csv(cross_validation_path)
    kv_df = pd.read_csv(key2value_path)
    spec_df = pd.read_csv(specific_path,index_col=0)
    ispec_df = pd.read_csv(ispecific_path,index_col=0)
    template_df = spec_df.reset_index().rename(columns={"index":"template"})
    itemplate_df = ispec_df.reset_index().rename(columns={"index":"template"})
    config = dict(
        cv_df = cv_df,
        kv_df = kv_df,
        template_df = template_df,
        select_num = select_num
    )

    allfp_mat = np.loadtxt(save_fp_path,delimiter=" ")
    indices = allfp_mat[:,0]

    def gen_paths(dir):
        paths = list(
            map(
                lambda x:os.path.join(dir,"ind0{}.npy".format(x)),
                range(n)
            )
        )
        # 去掉不存在的path
        paths = list(filter(os.path.exists,paths))
        return paths

    sub_data_paths = list(
        map(
            lambda x: os.path.join(data_dirs[0],"ind0{}.npy".format(x)),
            range(n)
        )
    )
    bounds = list(map(get_data_shape,sub_data_paths))
    biter = accumulate(bounds,operator.add)
        
    bound = next(biter)
    plist = reduce(split_func,indices,([],0,bound,[],biter))
    indices_slice_list = list(plist[3])
    last_one = list(plist[0])
    indices_slice_list.append(last_one)

    config["indices_slice_list"] = indices_slice_list

    fpdata_paths_mapped = map(
        gen_paths,fp_data_save_dirs
    )

    get_fpdf_mapped = map(
        lambda paths: cons_fpgraph_data(n,paths,**config),
        list(fpdata_paths_mapped)
    )

    def gen_save_path(dir):
        path = os.path.join(dir,"false_positive_superfamily.csv")
        return path
    
    def save_fpdf(*args):
        fpdf, save_path = args
        fpdf.to_csv(save_path,index=False)
    
    path_mapped = map(
        gen_save_path,fp_data_save_dirs
    )

    # list(
    #     map(
    #         lambda p: save_fpdf(*p),
    #         zip(
    #             list(get_fpdf_mapped),
    #             list(path_mapped)
    #         )
    #     )
    # )

    # 生成nodes和edges表格用于画图
    def gen_save_ndpaths(dir):
        return os.path.join(dir,"false_positive_nodes.csv")

    def gen_save_edpaths(dir):
        return os.path.join(dir,"false_positive_edges.csv")

    ndpath_mapped = map(
        gen_save_ndpaths,fp_data_save_dirs
    )

    edpath_mapped = map(
        gen_save_edpaths,fp_data_save_dirs
    )

    def save_nded(*args):
        fpdf, ndpath, edpath = args
        # 重新给fpdf编号
        # 把superfamily和sequence放在一块，统一编号
        # sequence排在superfamily后面
        # 要在重新编号前建立dataframe
        # 顶点是独立测试集的因此使用itemplate_df
        nd_df = itemplate_df.set_index("row num").loc[fpdf.loc[:, "row num"]].reset_index().rename(columns={"row num":"Id","template":"Label"})
        sfnd_df = kv_df.set_index("label").loc[fpdf.loc[:,"label"]].reset_index().rename(columns={"label":"Id","sf":"Label"})

        # 去重
        g = nd_df.groupby("Id")
        nd_df = g.apply(lambda df: df.loc[df.index[0],["Label"]])
        g = sfnd_df.groupby("Id")
        sfnd_df = g.apply(lambda df: df.loc[df.index[0],["Label"]])

        nd_df = nd_df.reset_index()
        sfnd_df = sfnd_df.reset_index()
    
        # 构建原Id对新Id的映射
        sf_size = sfnd_df.shape[0]
        sf_df = pd.DataFrame({"label": sfnd_df.loc[:,"Id"],"new_label":range(sf_size)})
        seq_df = pd.DataFrame({"row num": nd_df.loc[:,"Id"], "new_row_num":range(sf_size,sf_size+nd_df.shape[0])})
        new_sfid_ser = sf_df.set_index("label").loc[sfnd_df.loc[:,"Id"],"new_label"].reset_index(drop=True)
        new_seqid_ser = seq_df.set_index("row num").loc[nd_df.loc[:,"Id"],"new_row_num"].reset_index(drop=True)
        # 直接在原来的dataframe上进行替换
        sfnd_df.loc[:,"Id"] = new_sfid_ser
        nd_df.loc[:,"Id"] = new_seqid_ser

        # 把superfamily也作为顶点
        nd_df = pd.concat([sfnd_df,nd_df],axis=0)
        nd_df.loc[:,"timeset"] = None

        # 建立边表格
        ed_df = fpdf.loc[:,["row num","label"]].rename(columns={"row num":"Source","label": "Target"})
        # ed_df = ed_df.reset_index(drop=True)
        new_source_ser = seq_df.set_index("row num").loc[ed_df.loc[:,"Source"],"new_row_num"].reset_index(drop=True)
        new_target_ser = sf_df.set_index("label").loc[ed_df.loc[:,"Target"],"new_label"].reset_index(drop=True)
        ed_df.loc[:,"Source"] = new_source_ser
        ed_df.loc[:,"Target"] = new_target_ser
        ed_df.loc[:,["Type","Id","Label","timeset"]] = None
        # 已经提前处理过了
        # # 处理边的权重
        # # 保证不为负
        # score_ary = fpdf.loc[:,"score"].to_numpy(float)
        # if (fpdf.loc[:,"score"] < 0).any():
        #     min_score = score_ary.min()
        #     score_ary += np.abs(min_score)

        # absmin_score = np.abs(score_ary[score_ary > 0]).min()
        # # 加一个最小的值
        # # 保证不为0
        # epsilon = absmin_score * 10**(-round(absmin_score,0))
        # score_ary += epsilon
        # # 向边表格内添加权重
        # ed_df.loc[:,"Weight"] = score_ary
        ed_df.loc[:,"Weight"] = fpdf.loc[:,"score"]

        nd_df.to_csv(ndpath,index=False)
        ed_df.to_csv(edpath,index=False)


    list(
        map(
            lambda p:save_nded(*p),
            zip(
                list(get_fpdf_mapped),
                list(ndpath_mapped),
                list(edpath_mapped)
            )
        )
    )


if __name__ == "__main__":
    import numpy as np
    import pandas as pd
    from igraph import *
    import tensorflow as tf
    from functools import reduce
    from itertools import accumulate,groupby

    all_fold = 5
    # save_fgraph_data()
    # save_sfgraph_data()
    # call_fp_data(all_fold)
    save_fpgraph_data(all_fold)
    

# # %%
# import numpy as np

# n = 5
# i = 212

# pmapped = map(
#     lambda x:"feature/LTR/training/representation/ind0{}.npy".format(x), range(n)
# )

# mat_mapped = map(
#     np.load,
#     list(pmapped)
# )

# sel_mapped = map(
#     lambda m: m[i],
#     list(mat_mapped)
# )

# sel_argsort = map(
#     lambda v: np.argsort(-v,axis=0),
#     list(sel_mapped)
# )

# list(sel_argsort)

# a = np.ones((5,10))

# indices = np.where(a)
# indices = np.stack(indices,axis=1)
# indices

# df = pd.DataFrame({"A":[3,4,5,6],"B":[11,22,53,-12]})

# (df.loc[:, "A"] < 0).any()
# ones_mat = np.ones((10,5))
# iary = np.array(range(10)).reshape((-1,1))
# ones_mat * iary

# a = np.random.randint(0,2,(20,2),dtype=bool)
# a
# # %%
# import tensorflow as tf

# tf.map_fn(lambda v:tf.math.reduce_all(v,axis=0), a)
# # %%
# indices = np.where(a)
# # %%
# indices
# #%%
# imat = np.stack(indices,axis=0).T
# # %%
# imat
