# 与test_sequencew_identity一起使用
# 用于分析sequence identity的实验结果
# 通过图像的方式来展现
#%%
def gen_between_pair(iterable, r):
    def iter_func(i):
        return iterable[i:i+r]
    return map(iter_func,range(len(iterable)-r+1))

def common_config(specific_func):
    def config_func(diff_path,**kArgs):
        diffdf = pd.read_csv(diff_path)
        config = dict(
            x_labels = diffdf.columns[1:],
            x_ticks = diffdf.columns[1:].to_numpy(np.float32)
        )
        if kArgs is None:
            kArgs = config
        else:
            kArgs["x_labels"] = config["x_labels"]
            kArgs["x_ticks"] = config["x_ticks"]
        config = specific_func(**kArgs)
        return config
    return config_func

@common_config
def rfeature_importance_config(**kArgs):
    file_dir = "evaluation/feature_importance/sequence_identity"
    # seq_iden_ary = np.round(np.arange(0.4,1,0.05),2)
    seq_iden_ary = np.round(np.arange(0.25,1,0.05),2)
    feature_name_list = [
        "LinearSVM(HMM)",
        "LinearSVM(PSSM)",
        "DFG-S2C(HHblits)",
        "DFG-S2C(HHsearch)",
        "DFG-S2C(PSI-BLAST)",
        "DFG-S2C(DELTA-BLAST)"
    ]
    # kArgs["file_dir"] = file_dir
    file_path_list = list(
        map(
            lambda iden: os.path.join(file_dir,"seq_iden_{}.txt".format(iden)),
            seq_iden_ary
        )
    )
    kArgs["file_path_list"] = file_path_list
    kArgs["feature_name_list"] = feature_name_list
    return kArgs

@common_config
def rrconfig(**kArgs):
    # file_path = "output/lgb/lgb_diff_eval_20201022-13:49:59"
    file_path = "output/lgb/lgb_diff_eval_20201117-03:04:53"
    columns = ["ROC1","ROC50","AUC"]
    
    kArgs["file_path"] = file_path
    kArgs["columns"] = columns
    return kArgs

# file_dir 中存放了多个feature importance文件
def read_feature_importance(**kArgs):
    file_path_list = kArgs["file_path_list"]
    feature_name_list = kArgs["feature_name_list"]
    def read_single(fpath):
        # fpath = os.path.join(file_dir,file_name)
        ary = np.loadtxt(fpath,delimiter=" ",dtype=np.float32)
        return ary
    
    # 假设有m个feature importance文件，每个文件包含k个特征, 实验为n折交叉验证
    # 此处得到的是m * n * k大小的tensor
    ten = tf.stack(list(map(read_single,file_path_list)),axis=0)
    # 归一化
    # norm1 = tf.norm(ten,1,axis=2,keepdims=True)
    ord = 1
    norm_axis2 = tf.norm(ten,ord=ord,axis=2,keepdims=True)
    ten /= norm_axis2

    # 归一化后，取平均值
    # 得到m * k大小的矩阵
    dmat = tf.math.reduce_mean(ten,axis=1)
    ddf = pd.DataFrame(dmat,columns=feature_name_list)
    kArgs["fcedf"] = ddf
    return kArgs

def read_rocat(**kArgs):
    file_path = kArgs["file_path"]
    columns = kArgs["columns"]
    dmat = np.loadtxt(file_path,delimiter=",")
    ddf = pd.DataFrame(dmat,columns=columns)
    kArgs["rocdf"] = ddf
    return kArgs

def lines_config(**kArgs):
    if kArgs is None:
        kArgs = {}
    
    kArgs["edge_cs"] = ["#7B3434","#1122AC","#DDC011"]
    kArgs["uline_cs"] = ["#EE605F","#6EA6D5","#F7E997"]
    kArgs["markers"] = ["$\\boxtimes$","$\\otimes$","$\\Cap$"]
    kArgs["labels"] = ["ROC1","ROC50","AUC"]
    kArgs["fillstyle"] = "full"
    kArgs["linewidth"] = 1.5
    kArgs["markersize"] = 7
    kArgs["markeredgewidth"] = 0.4

    return kArgs

# 参考模板
# c="#1122AC",fillstyle='full',marker="$\\boxtimes$",
                #  markeredgecolor="#1122AC",linewidth=1.5,label="ROC1",markersize=7,markeredgewidth=0.4
def gen_params(i, **kArgs):
    params = dict(
        c = kArgs["edge_cs"][i],
        fillstyle=kArgs["fillstyle"],
        marker=kArgs["markers"][i],
        markeredgecolor=kArgs["edge_cs"][i],
        linewidth=kArgs["linewidth"],
        label=kArgs["labels"][i],
        markersize=kArgs["markersize"],
        markeredgewidth=kArgs["markeredgewidth"]
    )
    return params

# dict(c=None,marker="$\\vdash$",markersize=8,linestyle=(0, (3, 1, 1, 1, 1, 1)),linewidth=1.8,label=None)
def blines_config(**kArgs):
    if kArgs is None:
        kArgs = {}
    kArgs["edge_cs"] = ["#EA5B57","#1B6B9A","#88C786","#B280B8","#EBAB63","#009899"]
    kArgs["marker"] = "$\\vdash$"
    kArgs["markersize"] = "8"
    kArgs["linestyle"] = (0, (3, 1, 1, 1, 1, 1))
    kArgs["linewidth"] = 1.8

    return kArgs

def gen_box_params(i, **kArgs):
    params = dict(
        c = kArgs["edge_cs"][i],
        marker=kArgs["marker"],
        markersize=kArgs["markersize"],
        linestyle=kArgs["linestyle"],
        linewidth=kArgs["linewidth"],
        label=kArgs["labels"][i]
    )
    return params

def plot_line(ax, x):
    def plot_func(*args):
        y, params = args
        return ax.plot(x,y,**params)
    return plot_func

def fill_between(ax, x,ys, params_list):
    def between_func(*args):
        i, j = args
        return ax.fill_between(x,ys[i],ys[j],**(params_list[i]))
    return between_func

def plot_bls(**kArgs):
    fcedf = kArgs["fcedf"]
    rocdf = kArgs["rocdf"]
    
    x_ticks = kArgs["x_ticks"]
    x_labels = kArgs["x_labels"]

    fig, axs = pl.subplots(1,2, sharex=False,figsize=(15,7),dpi=1200)
    pl.subplots_adjust(wspace =0.12, hspace =0.02)
    ax0, ax1 = axs

    # 字体设置
    pl.rcParams.update({
        'font.family':'sans-serif',
        'font.sans-serif':['Arial Unicode MS'],
    })
    ax0.tick_params(direction="in")
    ax1.tick_params(direction="in")

    # 先绘制roc1, roc50, auc变化的曲线图
    # 进行相关设置
    n_lines = 3
    config = lines_config()
    plot_config_list = list(map(lambda i: gen_params(i,**config), range(n_lines)))

    legend_loc = (0.01,0.3)

    # 获取数据
    # 变成 3 * m大小
    # 这里的3指roc1, roc50, auc这3种值
    # m指进行实验的次数
    data = rocdf.to_numpy(dtype=np.float32).T
    list(map(lambda p: plot_line(ax0,x_ticks)(*p),zip(data,plot_config_list)))
    # ax0.legend(loc="center left")
    ax0.legend(loc=legend_loc)
    a, b = x_ticks[0], x_ticks[-1]
    ax0.set_xlim(a,b)
    # 0.93,0.955,0.0025
    boundary,end = np.round(data.min(),1), np.round(data.max(),4)
    step = np.round((end-boundary) / data.shape[1],3)

    ax0.set_ylim(boundary,end+step)
    y_ticks = np.arange(boundary,end+2*step,step)
    ax0.set_xticks(x_ticks)
    ax0.set_yticks(y_ticks)
    # 对线条之间进行填充
    between_cs = ["#EE605F","#6EA6D5","#F7E997"]
    bconfigs = [dict(color=x) for x in between_cs]
    ys = [boundary] + data.tolist()
    list(map(lambda p: fill_between(ax0,x_ticks,ys,bconfigs)(*p), gen_between_pair(range(4),2)))

    # 绘制盒图
    boxconf_dict = dict(color="#B22222",linewidth=1.2)
    whiskerconf_dict = dict(color="#B22222",linewidth=1.2)
    medianconf_dict = dict(color="#B22222",linewidth=1.9)
    caps_dict = dict(color="#B22222",linewidth=1.2)

    # 同样转成6 * m大小
    # 这里6指基方法的个数
    # m为试验次数
    fdata = fcedf.to_numpy(dtype=np.float32).T
    bplot = ax1.boxplot(fdata,boxprops=boxconf_dict,
        capprops=caps_dict,whiskerprops=whiskerconf_dict,medianprops=medianconf_dict)

    # 绘制盒图中的曲线
    ax1.set_xticklabels(x_labels)
    bx = np.arange(1,x_ticks.shape[0]+1,1)
    config = blines_config(labels = kArgs["feature_name_list"])
    bplot_config_list = list(map(lambda i: gen_box_params(i,**config), range(len(kArgs["feature_name_list"]))))
    list(map(lambda p: plot_line(ax1,bx)(*p),zip(fdata,bplot_config_list)))
    ax1.legend(loc="center left")
    ax1.legend(loc=legend_loc)

    boundary, end = np.round(fdata.min(),1), np.round(fdata.max(),4)
    step =  np.round((end-boundary) / data.shape[1],2)
    ax1.set_ylim(boundary,end+step)
    y_ticks = np.arange(boundary,end+2*step,step)
    ax1.set_yticks(y_ticks)

# 分析数据集中远同源与非远同源的比例
def fasta2dict(level, fas_path):
    from Bio.SeqIO import parse
    
    if level == "fold":
        n_pos = 1
    elif level == "sf":
        n_pos = 2
    elif level == "family":
        n_pos = 3
    else:
        assert level in ["fold", "sf", "family"]
    
    # 把description转为protein对应的type，比如superfamily
    def des2type(des):
        import re

        searcher = re.compile(r"\w+(\.\w+){%d}" % n_pos)
        type = searcher.search(des)
        return type.group()
    
    def recoder2sfseq(r):
        return des2type(r.description), r.name

    with open(fas_path,"r") as h:
        sfseq_list = list(map(recoder2sfseq,parse(h,"fasta")))
    
    sfdict = {}

    def cons_sfdict(*args):
        sf, seq = args
        if sfdict.get(sf) is None:
            sfdict[sf] = []
        sfdict[sf].append(seq)
    
    list(map(lambda p: cons_sfdict(*p), sfseq_list))
    return sfdict

# 得到用于绘制饼图的数据
# 假设n_sf为superfamily的个数
# 一个大小为n_sf * n_sf大小的矩阵
def dict2pichartData(sfdict,kv_df):
    def get_length(*args):
        sf, seq_list = args
        return sf, len(seq_list)
    sz_df = pd.DataFrame(list(map(lambda p: get_length(*p),sfdict.items())),columns=["sf","size"])
    vdf = kv_df.groupby("sf").apply(lambda d:d.loc[d.index[0]])
    sz_df = sz_df.join(vdf.loc[:,["sf","label"]].set_index("sf"),on="sf")
    # 转成向量方便计算
    sz_ary = sz_df.loc[:,"size"].to_numpy(np.float32).reshape((-1,1))
    # 这样计算是为了得到不同superfamily组合的结果
    pimat = (sz_ary @ sz_ary.T)
    # 同一组要减去重复的结果
    pimat -= np.eye(sz_ary.shape[0]) * sz_ary
    # 取上三角是因为，要减去不同组合之间的重复结果
    pimat = np.triu(pimat)

    # 计算属于远同源的pairwise个数
    h = np.trace(pimat)
    # 所有pairwise个数
    s = np.sum(pimat)
    # nh表示非远同源
    nh = s - h
    return h, nh, s


# 这里绘制饼图时，要用dataset/scop_scope/ori中的数据
# 原因是因为，我们要描绘的时独立测试机与训练集之间的

#%%
if __name__ == "__main__":
    import matplotlib as mpl
    import matplotlib.pylab as pl
    import matplotlib.cm as cm
    import os
    import pandas as pd
    import numpy as np
    import tensorflow as tf

    the_diff_path = "dataset/scop_scope/csv/sequence_identity.csv"
    config = rfeature_importance_config(the_diff_path)
    config = rrconfig(the_diff_path,**config)
    config = read_feature_importance(**config)
    config = read_rocat(**config)
    # print(config)
    plot_bls(**config)
# %%
import os

# %%
