# 进行k折交叉验证

import os
import subprocess as sp
from sklearn.model_selection import KFold
from Bio.SeqIO import parse, write
import numpy as np
from six.moves import cPickle
import pandas as pd
from sklearn.metrics import auc, roc_curve, roc_auc_score
from sklearn.preprocessing import label_binarize

class Run(object):
    def __init__(self, benchmarkDatasetPath=None):
        self.benchmarkDatasetPath = benchmarkDatasetPath
    
    def spilt_n_folds(self,datasetPath=None,n=5,resultPath=None):
        if datasetPath is None:
            datasetPath = self.benchmarkDatasetPath
        if resultPath is None:
            datasetDir = os.path.dirname(datasetPath)
            resultPath = os.path.join(datasetDir,"cross_validation")
            os.mkdir(resultPath)
        
        seqName_l = []
        with open(datasetPath,"r") as h:
            for record in parse(h,"fasta"):
                name = record.name
                seqName_l.append(name)
        
        sName_ary = np.array(seqName_l)
        sName2fNum = {}

        kf = KFold(shuffle=True,random_state=43)
        fold_num = 0
        for _, query_index in kf.split(sName_ary):
            for i in query_index:
                qSName = sName_ary[i]
                if sName2fNum.get(qSName) is None:
                    sName2fNum[qSName] = fold_num
            fold_num += 1

        fNum2recod = {}
        with open(datasetPath,"r") as h:
            for record in parse(h,"fasta"):
                name = record.name
                fNum = sName2fNum[name]
                if fNum2recod.get(fNum) is None:
                    fNum2recod[fNum] = []
                fNum2recod[fNum].append(record)
        
        for i in range(n):
            datasetName = os.path.basename(datasetPath)
            resultFilePath = os.path.join(resultPath,"{}_foldNo{}".format(datasetName,i))
            with open(resultFilePath,"w") as h:
                write(fNum2recod[i],h,"fasta")
        
        snAryPath = os.path.join(resultPath,"sequence_name_array")
        sn2fnPath = os.path.join(resultPath,"sequence_name2fold_num")

        with open(snAryPath,"wb") as h,\
            open(sn2fnPath,"wb") as h1:
            np.save(h,sName_ary)
            cPickle.dump(sName2fNum,h1)
    
    # 根据已分好的结果，进行交叉验证，目前数据集被分为5份，结果保存在dataset/benchmark/cross/fold_num2sequence_name.csv中
    # 通过不同的列来区分这5份子集
    # 实验的时候调用pre_train_model中的cross_validation进行实验
    # def experiment(self, pre_train_model)

    # 计算roc1, roc50这样的值
    # srch_df也是检索结果，在use_srch为True的情况下，如果提供了srch_df那么，就直接计算srch_df的rocat
    def roc_at(self, level=1, cv_df=None,spec_df=None, kv_dict=None, ltr_npy=None, srch_df=None, use_srch=False):
        def cv(cv_df):
            for fold_num in cv_df.columns:
                start = 0
                for queryName in cv_df.loc[:,fold_num]:
                    if pd.isna(queryName):
                        continue
                    # s_df = spec_df.reset_index().assign(query=queryName).rename(columns={"index":"template"})
                    # s_df = s_df.loc[~s_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
                    # pairwise_df = s_df.loc[:,["query","template"]]

                    # 同样通过searched result过滤一下
                    s_df = srch_df.loc[srch_df.loc[:,"query"] == queryName]
                    s_df = s_df.loc[~s_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
                    pairwise_df = s_df.loc[:,["query","template"]]

                    size = pairwise_df.shape[0]
                    rank_npy = ltr_npy[start:start+size]
                    pairwise_df = pairwise_df.assign(score=rank_npy)
                    start += size
                    yield pairwise_df
        
        def cal_roc(pairwise_df):
            # print(pairwise_df)
            pairwise_df = pairwise_df.loc[~pd.isna(pairwise_df.loc[:,"template"])]
            if pairwise_df.shape[0] == 0:
                return 0.0
            pairwise_df = pairwise_df.sort_values(by="score",ascending=False)
            fold_type_df = \
                pairwise_df.loc[:,["query","template"]].apply(lambda d: d.apply(lambda x: kv_dict[x]))
                        
            label_ss = (fold_type_df.loc[:,"query"] == fold_type_df.loc[:,"template"])
            # 统计所有的true positive
            tp = pairwise_df.loc[label_ss].shape[0]
            # 统计level之前的fp, 以及第level个false positive所在的row index
            # 同时计算area和height
            # area是每个fp之前height的个数之和
            # height是level之前tp的个数

            l_ser = label_ss.loc[:level-1]
            fp = l_ser.value_count().loc[False]

            if fp == 0:
                return 1.0
            
            p_df = pairwise_df.loc[:level-1]
            fp_idx = p_df.index[~l_ser]

            if fp_idx.shape[0] > level:
                level_idx = fp_idx[level-1]
            else:
                level_idx = fp_idx[fp_idx.shape[0]-1]


            # 这种写法不对
            # fp_idx = pairwise_df.index[~label_ss]
            # if fp_idx.shape[0] > level:
            #     fp = level
            #     level_idx = fp_idx[level-1] # 下标从0开始
            # elif fp_idx.shape[0] != 0:
            #     fp = fp_idx.shape[0]
            #     level_idx = fp_idx[fp-1] # 取最后一个
            # else:
            #     return 1.0
                        
            # level_idx表示的是Series中的index
            level_ss = label_ss.loc[:level_idx] # 这种取法包括第level_idx行也在结果中
            sum_ss = level_ss.cumsum(axis=0)
            area_idx = fp_idx[:level] # level前的所有fp所在位置，这种取法把第level也能包含进去（下标从0开始）
            area = sum_ss.loc[area_idx].sum()
            if tp != 0 and fp != 0:
                rocat = area / (tp * fp)
            elif tp != 0 and fp == 0:
                rocat = 1.0
            elif tp == 0 and fp != 0:
                rocat = 0.0
            
            return rocat
        
        
        if use_srch and srch_df is not None:
            p_df = srch_df.loc[:,["query","template","prob"]] # 以prob值为基准来排序
            p_df = p_df.rename(columns={"prob":"score"})
            # print(p_df)
            pDf_g = p_df.groupby("query")
            roc_ss = pDf_g.apply(cal_roc)
            final_roc = roc_ss.mean()
        else:
            # roc_list = []
            # for pairwise_df in cv(cv_df):
            #     rocat = cal_roc(pairwise_df)
                
            #     roc_list.append(rocat)
            roc_ss = pd.Series([cal_roc(d) for d in cv(cv_df)])
            final_roc = roc_ss.mean()
        
        return final_roc
    
    def auc_mu(self, fold_num, cv_df, spec_df, vspec_df, kv_df, clf, rep, vrep=None):
        valid_df = vspec_df.join(kv_df.set_index("template"))
        vlabel = valid_df.loc[:,"label"].to_numpy(np.int64)
        template_ss = pd.Series(spec_df.index,name="template")

        train_df = spec_df.loc[template_ss.loc[~template_ss.isin(cv_df.loc[:,fold_num])]]
        # print(train_df)
        train_df = train_df.join(kv_df.set_index("template"))
        # print(train_df)
        # 重新给superfamily编号
        k_df = train_df.groupby("sf").apply(lambda x: x.iloc[0,0]).reset_index()
        # print(k_df)
        k_df = k_df.reset_index().loc[:,["sf","index"]]
        # print(k_df)
        train_df = train_df.join(k_df.set_index("sf"),on="sf")
        train_df = train_df.rename(columns={"index":"new label"})
        # classes = [x for x in train_df.groupby("new label").groups.keys()]

        train_df = train_df.sort_values(by="new label")
        # train_df
        classes = [x for x in train_df.groupby("label").groups.keys()]

        if vrep is not None:
            # 评估在验证集上的性能
            y_pred = clf.decision_function(vrep)
            am = self.cal_auc_mu(y_pred,vlabel,classes)
            # print("验证集auc_mu: {}".format(am))
            # cli_print = "printf '验证集auc_mu: {}\n'".format(am)
            # sp.call(cli_print,shell=True)
        
        # cli_print = "printf '使用auc_mu评估模型在测试集上的性能\n'"
        # sp.call(cli_print,shell=True)

        # 评估在测试集上的性能
        test_df = spec_df.loc[template_ss.loc[template_ss.isin(cv_df.loc[:,fold_num])]]
        test_df = test_df.join(kv_df.set_index("template"))
        # test_df
        tlabel = test_df.loc[:,"label"].to_numpy(np.int64)
        idx = test_df.loc[:, "row num"]
        X_test = rep[idx]
        y_pred = clf.decision_function(X_test)
        am = self.cal_auc_mu(y_pred, tlabel, classes)
        cli_print = "printf '测试集auc_mu: {}\n'".format(am)
        sp.call(cli_print,shell=True)        


    def cal_auc_mu(self, y_pred, y_true, classes):
        y_true = label_binarize(y_true, classes=classes)
        # fpr = dict()
        # tpr = dict()
        # roc_auc = []
        # for i in range(len(classes)):
        #     fpr[i], tpr[i], _ = roc_curve(y_true[:, i], y_pred[:, i])
        #     if not np.isnan(tpr[i]).any():
        #         try:
        #             roc_auc.append(auc(fpr[i], tpr[i]))
        #         except:
        #             if 1 in tpr[i]:
        #                 roc_auc.append(1.)
        #             else:
        #                 roc_auc.append(0.)
        # auc_score = np.array(roc_auc).mean()
        y_pred = y_pred.reshape((-1,))
        y_true = y_true.reshape((-1,))

        auc_score = roc_auc_score(y_true,y_pred)

        return auc_score

    # 当vrep为None
    # 表示对测试集性能进行评估
    # level 可以是int类型也可以是list类型
    def roc_mu_at(self, level, fold_num, cv_df, spec_df, vspec_df, kv_df, clf, rep, vrep=None):
        valid_df = vspec_df.join(kv_df.set_index("template"))
        vlabel = valid_df.loc[:,"label"].to_numpy(np.int64)
        template_ss = pd.Series(spec_df.index,name="template")
        train_df = spec_df.loc[template_ss.loc[~template_ss.isin(cv_df.loc[:,fold_num])]]
        # print(train_df)
        train_df = train_df.join(kv_df.set_index("template"))
        # print(train_df)
        # 重新给superfamily编号
        k_df = train_df.groupby("sf").apply(lambda x: x.iloc[0,0]).reset_index()
        # print(k_df)
        k_df = k_df.reset_index().loc[:,["sf","index"]]
        # print(k_df)
        train_df = train_df.join(k_df.set_index("sf"),on="sf")
        train_df = train_df.rename(columns={"index":"new label"})
        # classes = [x for x in train_df.groupby("new label").groups.keys()]
        # train_df = train_df.sort_values(by="new label")
        # train_df
        classes = [x for x in train_df.groupby("label").groups.keys()]

        if vrep is not None:
            y_pred = clf.decision_function(vrep)
            y_true = label_binarize(vlabel, classes=classes)
        else:
            test_df = spec_df.loc[template_ss.loc[template_ss.isin(cv_df.loc[:,fold_num])]]
            idx = test_df.loc[:, "row num"]
            X_test = rep[idx]
            test_df = test_df.join(kv_df.set_index("template"))
            tlabel = test_df.loc[:,"label"].to_numpy(np.int64)
            y_pred = clf.decision_function(X_test)
            y_true = label_binarize(tlabel, classes=classes)

        indices = np.argsort(y_pred,1)
        n, _ = indices.shape
        roc_at_dict = {}

        # print(y_true.shape)
        # print(n)
        for j in range(n):
            yt = np.flip(y_true[j, indices[j]])
            # if not yt.astype(bool).any():
            #     continue
            # print(yt[0])
            if type(level) == list:
                for lv in level:
                    if roc_at_dict.get(lv) is None:
                        roc_at_dict[lv] = {}
                    roc_at_dict[lv][j] = self.cal_roc_mu_at(lv, yt)
            else:
                roc_at_dict[j] = self.cal_roc_mu_at(level, yt)

            # print(roc_at_dict[1][j],roc_at_dict[50][j])
        # 这里还要根据level的类型，统计最终的结果            
        if type(level) == list:
            score = [
                np.array(list(d.values())).mean() \
                    for d in roc_at_dict.values()
            ]
        else:
            score = [np.array(list(roc_at_dict.values())).mean()]
        # print(len(roc_at_dict[1]))
        return score
    
    def acc(self, fold_num, cv_df, spec_df, vspec_df, kv_df, clf, rep, vrep=None):
        valid_df = vspec_df.join(kv_df.set_index("template"))
        vlabel = valid_df.loc[:,"label"].to_numpy(np.int64)
        template_ss = pd.Series(spec_df.index,name="template")
        train_df = spec_df.loc[template_ss.loc[~template_ss.isin(cv_df.loc[:,fold_num])]]
        # print(train_df)
        train_df = train_df.join(kv_df.set_index("template"))
        # print(train_df)
        # 重新给superfamily编号
        k_df = train_df.groupby("sf").apply(lambda x: x.iloc[0,0]).reset_index()
        # print(k_df)
        k_df = k_df.reset_index().loc[:,["sf","index"]]
        # print(k_df)
        train_df = train_df.join(k_df.set_index("sf"),on="sf")
        train_df = train_df.rename(columns={"index":"new label"})
        # classes = [x for x in train_df.groupby("new label").groups.keys()]
        # train_df = train_df.sort_values(by="new label")
        # train_df
        classes = [x for x in train_df.groupby("label").groups.keys()]

        if vrep is not None:
            y_pred = clf.predict(vrep)
        else:
            test_df = spec_df.loc[template_ss.loc[template_ss.isin(cv_df.loc[:,fold_num])]]
            idx = test_df.loc[:, "row num"]
            X_test = rep[idx]
            test_df = test_df.join(kv_df.set_index("template"))
            tlabel = test_df.loc[:,"label"].to_numpy(np.int64)
            y_pred = clf.predict(X_test)
        
        # 将y_pred转为原来的label
        yp_df = pd.DataFrame(y_pred,columns=["new label"])
        yp_df = yp_df.join(train_df.set_index("new label"),on="new abel")
        y_pred = yp_df.loc[:,"label"]


    # srch_df包含query, template, prob三列
    # srch_ary每一行代表一种排序结果
    # srch_df,kv_i_df符合交叉验证的划分结果
    # kv_i_df用于扩展当前排序列表
    def srch_roc_at(self, level, kv_df, kv_i_df, srch_df, srch_ary=None,expand_flag=True):
        # 当srch_ary为None时，表示srch_df就是要评估的对象
        if srch_ary is None:
            srch_df = srch_df.join(kv_df.set_index("template"),on="query")
            srch_df = srch_df.join(kv_df.set_index("template"),on="template",lsuffix="_query",rsuffix="_template")
            srch_df = srch_df.sort_values(by="prob", ascending=False)
            label = srch_df.loc[:, "sf_query"] == srch_df.loc[:, "sf_template"]

            if expand_flag:
                query_sf = srch_df.loc[srch_df.index[0],"sf_query"]
                idx = kv_i_df.loc[:,"template"].isin(srch_df.loc[:,"template"])
                if idx.any():
                    ex_label = (query_sf == kv_i_df.loc[~idx,"sf"])
                    label = pd.concat([label, ex_label],axis=0)
            y_true = label.to_numpy(int)
        
        # 当srch_ary不为None时，则表示srch_ary时要评估的对象
        # 此时srch_df包含query, template两列
        # 方便生成y_true
        else:
            # 要拓展排序列表，把属于这一折的所有template都要囊括进来
            # print(srch_df)
            label = srch_df.loc[:, "sf_query"] == srch_df.loc[:, "sf_template"]
            y_true = label.to_numpy(int)

            # 根据排序结果来，把y_true进行排序
            idx = np.flip(np.argsort(srch_ary))
            y_true = y_true[idx]

        return self.cal_roc_mu_at(level,y_true)      

    
    def cal_roc_mu_at(self, level, y_true):
        # 统计所有的true positive
        tp = np.count_nonzero(y_true)
        # print(tp)
        # 统计level个false positive之前的tp之和
        all_fp = np.count_nonzero(~y_true.astype(bool))
        if all_fp < level:
            #如果不足就填充False，因为，没检索出来
            fp = all_fp
            yt_level = y_true
        else:
            df = pd.DataFrame(y_true.astype(bool),columns=["label"])
            fp_index = df.loc[~df.loc[:,"label"]].index[level-1]
            yt_level = df.loc[:fp_index].to_numpy(int).reshape((-1,))

        fp = np.count_nonzero(~yt_level.astype(bool))
        # print(fp)
        
        # print(yt_level)
        cumsum_yt_level = np.cumsum(yt_level)
        area = cumsum_yt_level[~yt_level.astype(bool)].sum()
        # print(area)

        if tp != 0 and fp != 0:
            roc_at_score = area / (tp * fp)
        elif tp != 0 and fp == 0:
            roc_at_score = 1.0
        elif tp == 0 and fp != 0:
            roc_at_score = 0.0
        return roc_at_score
    
    # 测试数据集能达到的理想roc_at值是多少
    def perfect_roc_at(self, level, fold_num, cv_df, spec_df, vspec_df, kv_df, rep, vrep=None):
        valid_df = vspec_df.join(kv_df.set_index("template"))
        vlabel = valid_df.loc[:,"label"].to_numpy(np.int64)
        template_ss = pd.Series(spec_df.index,name="template")
        train_df = spec_df.loc[template_ss.loc[~template_ss.isin(cv_df.loc[:,fold_num])]]
        # print(train_df)
        train_df = train_df.join(kv_df.set_index("template"))
        # print(train_df)
        # 重新给superfamily编号
        k_df = train_df.groupby("sf").apply(lambda x: x.iloc[0,0]).reset_index()
        # print(k_df)
        k_df = k_df.reset_index().loc[:,["sf","index"]]
        # print(k_df)
        train_df = train_df.join(k_df.set_index("sf"),on="sf")
        train_df = train_df.rename(columns={"index":"new label"})
        # classes = [x for x in train_df.groupby("new label").groups.keys()]
        # train_df = train_df.sort_values(by="new label")
        # train_df
        classes = [x for x in train_df.groupby("label").groups.keys()]

        if vrep is not None:
            y_true = label_binarize(vlabel, classes=classes)
        else:
            test_df = spec_df.loc[template_ss.loc[template_ss.isin(cv_df.loc[:,fold_num])]]
            idx = test_df.loc[:, "row num"]
            X_test = rep[idx]
            test_df = test_df.join(kv_df.set_index("template"))
            tlabel = test_df.loc[:,"label"].to_numpy(np.int64)
            y_true = label_binarize(tlabel, classes=classes)

        # # 理想化的结果是，所有的序列都能预测正确
        y_true = np.sort(y_true,axis=1)
        n, _ = y_true.shape

        # y_true = np.sort(y_true, axis=0)
        # _,n = y_true.shape
        roc_at_dict = {}

        for j in range(n):
            yt = np.flip(y_true[j])
            # if not yt.astype(bool).any():
            #     continue
            # print(yt[0])
            # yt = np.flip(y_true[:,j])

            if type(level) == list:
                for lv in level:
                    if roc_at_dict.get(lv) is None:
                        roc_at_dict[lv] = {}
                    roc_at_dict[lv][j] = self.cal_roc_mu_at(lv, yt)
            else:
                roc_at_dict[j] = self.cal_roc_mu_at(level, yt)
        # 这里还要根据level的类型，统计最终的结果            
        if type(level) == list:
            score = [
                np.array(list(d.values())).mean() \
                    for d in roc_at_dict.values()
            ]
        else:
            score = [np.array(list(roc_at_dict.values())).mean()]
        return score  
    
    def norms(self, mat, ords, axis=1):
        norm_list = [
            np.linalg.norm(
                mat, x, axis=axis,keepdims=True
            ) \
                for x in ords
        ]
        norms = np.concatenate(norm_list,axis=1)
        return norms

