import os
import re
import subprocess as sp
import pandas as pd
import numpy as np
from six.moves import cPickle
from itertools import combinations
import tensorflow as tf

def getFileNameWithoutSurfix(psName):
    rv_psN = psName[::-1]
    name_size = len(psName)
    surfixParser = re.compile(r"\.")
    surfix = ""
    s_r = surfixParser.search(rv_psN)
    if s_r is not None:
        surfix_start = s_r.start()
        name = psName[:name_size-surfix_start-1]
        surfix = psName[name_size-surfix_start:]
    else:
        name = psName
    return name, surfix

def ac_calculation(col, LG):
    c_mean = col.mean()
    c_size = col.shape[0]
    c1 = (col - c_mean).to_numpy(dtype=float)
    def gen(c,LG):
        for lg in range(1,LG+1):
            p = c[0:c_size-lg] * c[lg:c_size]
            yield (p.sum() / (c_size - lg))
    r = list(gen(c1,LG))
    return r
def cc_calculation(col1,col2, LG):
    c_size = col1.shape[0]
    cm1 = col1.mean()
    cm2 = col2.mean()
    c1 = (col1 - cm1).to_numpy(dtype=float)
    c2 = (col2 - cm2).to_numpy(dtype=float)
    def gen(c1,c2,LG):
        for lg in range(1,LG+1):
            p = c1[0:c_size-lg] * c2[lg:c_size]
            yield (p.sum() / (c_size - lg))
    r = list(gen(c1,c2,LG))
    return r

# 矩阵运算，快速版本
def acc_transform(ary,LG):
    # 行归一化
    r,c = ary.shape
    # ary = pssm_df.to_numpy()
    l1norm = np.linalg.norm(ary,1,1)
    ary = ary / l1norm.reshape((-1,1))
    mean = np.average(ary,axis=0)
    ary -= mean
    # 计算ac
    # 计算cc
    ac_l = []
    cc_l = []
    for lg in range(1,LG+1):
        ac_l.append(1 / (r - lg) * np.sum(ary[:r-lg] * ary[lg:],axis=0))
        tmp = ary[:r-lg].T @ ary[lg:]
        # 去掉对角线上的值   
        cc_l.append(1 / (r - lg) * tmp[~np.eye(tmp.shape[0],dtype=bool)])
    ac_ary = np.hstack(ac_l).astype(np.float32)
    cc_ary = np.hstack(cc_l).astype(np.float32)

    return np.hstack([ac_ary,cc_ary]).astype(np.float32)

def top_n_gram_trans(mat, n, L):
    size = [L]*n
    gidx = np.flip(np.argsort(mat,1),1)
    if n == 1:
        gi_ser = pd.Series(gidx[:,:n].reshape((-1,)))
    else:
        gi_ser = pd.Series([tuple(p) for p in gidx[:,:n]])

    vc = gi_ser.value_counts()
    # print(vc)
    idx = vc.index
    counts = vc.to_numpy()

    if n == 1:
        grams = np.zeros(size)
        grams[idx] = counts
    else:
        idx = np.array([list(p) for p in vc.index]).T
        g_mat = np.zeros(size)
        g_mat[tuple(idx)] = counts
        grams = np.hstack(g_mat)

    all_size = grams.shape[0] # 表示所有n-gram的数量，比如2-gram是400,3-gram是8000
    grams /= all_size
    return grams.astype(np.float32)

# 用于提取特征
# -j 3 -h 0.001
# 以上是blastpgp的参数，j表示max rounds，h表示e-value
# blastpgp是更早版本里的一个工具，现在的blast+用psiblast代替
# 上述的两个参数相当于psiblast里的num_iterations和evalue
class PsiBlast(object):
    def __init__(self, db=None, query=None, out=None, evalue=1e-3, outfmt="6", num_threads=10,num_iterations=5):
        self.db=db
        self.query=query
        self.out = out
        self.evalue=evalue
        self.outfmt=outfmt
        self.num_threads=num_threads
        self.num_iterations = num_iterations
    
    # set_params用于重新设定参数
    # 这是一个dict类型
    def use_psiblast(self, set_params=None):
        def gen_psiblast_params(db,query,out,evalue,outfmt,num_threads,num_iterations):
            params = " psiblast -db {} -query {} -out {} \
                -evalue {} -outfmt {} -num_threads {} \
                -num_iterations {}".format(
                    db,query,out,evalue,outfmt,num_threads,
                    num_iterations
                )
            return params

        if set_params is None:
            set_params = {
                "db": self.db,
                "query": self.query,
                "out": self.out,
                "evalue": self.evalue,
                "outfmt": self.outfmt,
                "num_threads": self.num_threads,
                "num_iterations":self.num_iterations
            }
        params = gen_psiblast_params(**set_params)

        # cmd = "psiblast {}".format(params)
        cmd = params
        err_code = sp.call(cmd,shell=True)
        if err_code != 0:
            print("err_code: {}".format(err_code))
    
    # psiblast -in_pssm $query_name.pssm -db scope-95-2.06 –out $query_name.txt  -outfmt 5 -evalue 10 -num_iterations 5
    def search_by_pssm(self, set_params):
        def gen_psiblast_params(in_pssm,db,out,outfmt,evalue,num_threads, num_iterations):
            params = "psiblast -in_pssm {} -db {} -out {} -outfmt {} -evalue {} \
                -num_threads {} -num_iterations {}".format(
                    in_pssm,db,out,outfmt,evalue,num_threads, num_iterations
                )
            return params
        cmd = gen_psiblast_params(**set_params)
        sp.call(cmd, shell=True)


    def deltablast_checkpoints(self, set_params):
        def gen_deltablast_params(query,db,num_alignments,out_pssm, num_threads,rpsdb):
            params = "-query {} -db {} \
                 -num_alignments {} -out_pssm {} -save_pssm_after_last_round \
                     -num_threads {} -rpsdb {}".format(
                     query,db,num_alignments,out_pssm,num_threads,
                     rpsdb
                 )
            return params
        cmd = "deltablast {}".format(gen_deltablast_params(**set_params))
        sp.call(cmd,shell=True)
    
    def use_makeblastdb(self, set_params):
        def gen_makeblastdb_params(infile, input_type, dbtype, out):
            params = "-in {} -input_type {} -dbtype {} -out {}".format(
                infile,input_type,dbtype,out
            )
            return params
        
        params = gen_makeblastdb_params(**set_params)
        cmd = "makeblastdb {}".format(params)

        err_code = sp.call(cmd,shell=True)
        if err_code != 0:
            print("error code is {}".format(err_code))
    
    #把检索结果解析成一个dict文件
    def parsing_searched_result(self, searchedFilePath, parsingDict=None):
        start = re.compile(r"# [0-9]+ (hit|hits) found")
        contentStart = re.compile(r"^[^#\s]")
        end = re.compile(r"# BLAST processed [0-9]+ (query|queries)")


        getContentFlag = False
        content_l = []
        with open(searchedFilePath,"r") as h:
            for line in h:
                if not getContentFlag:
                    start_r = start.search(line)
                    if start_r is not None:
                        content_l = []
                        getContentFlag = True
                else:
                    end_r = end.search(line)
                    if end_r is not None:
                        break
                    cr = contentStart.search(line)
                    if cr is not None:
                        content_l.append(line)
                    else:
                        getContentFlag = False

        contentParser = re.compile(r"[ \t\n]")
        if parsingDict is None:
            parsingDict = {"query":[],"template":[],
                "percentage of identical matches":[],
                "expect value":[],"bit score":[]}
        q_num = len(parsingDict["query"])
        for c in content_l:
            parsing_result = contentParser.split(c)
            parsingDict["query"].append(parsing_result[0])
            parsingDict["template"].append(parsing_result[1])
            parsingDict["percentage of identical matches"].append(parsing_result[2])
            parsingDict["expect value"].append(parsing_result[3])
            parsingDict["bit score"].append(parsing_result[4])
        # 在没有反馈的情况下也要把query填充一个，然后其余部分为none或者pd.NA
        new_qN = len(parsingDict["query"])
        if q_num == new_qN:
            # 重新解析文件，找到query的name
            queryMatcher = re.compile(r"# Query:\s(?P<query>[a-z0-9A-Z_\.]+)")

            with open(searchedFilePath,"r") as h:
                for line in h:
                    query_r = queryMatcher.search(line)
                    if query_r is not None:
                        queryName = query_r.group("query")
            parsingDict["query"].append(queryName)
            parsingDict["template"].append(pd.NA)
            parsingDict["percentage of identical matches"].append(pd.NA)
            parsingDict["expect value"].append(pd.NA)
            parsingDict["bit score"].append(pd.NA)
        return parsingDict
    
    # psiblast -query {} -db {} -num_iterations {} -evalue {} -out_ascii_pssm {}
    def extract_pssm(self, set_params):
        def gen_cmd_params(query,db,num_iteration,evalue,out_ascii_pssm, num_threads):
            params = "-query {} -db {} \
                 -num_iterations {} -evalue {} -out_ascii_pssm {} -num_threads {}".format(
                     query,db,num_iteration,evalue,out_ascii_pssm,num_threads
                 )
            return params
        cmd = "psiblast {}".format(gen_cmd_params(**set_params))
        # print(cmd) # 检查输出
        sp.call(cmd,shell=True)

    def extract_pssm_ex(self, set_params):
        def gen_cmd_params(query,db,num_alignments,out_pssm, num_threads):
            params = "-query {} -db {} \
                 -num_alignments {} -out_pssm {} -save_pssm_after_last_round \
                     -num_threads {}".format(
                     query,db,num_alignments,out_pssm,num_threads
                 )
            return params
        cmd = "psiblast {}".format(gen_cmd_params(**set_params))
        # print(cmd) # 检查输出
        sp.call(cmd,shell=True)   
    
    # 把单个fasta文件转为pssm
    def fasta2pssm(self, fastaFilePath, blosum62FilePath, pssm_csvFormat_path=None):
        from Bio.SeqIO import parse

        b_df = pd.read_csv(blosum62FilePath,index_col=0)

        with open(fastaFilePath,"r") as h:
            for record in parse(h,"fasta"):
                seq = record.seq
        s_list = list(str(seq).upper())
        pssm_df = b_df.loc[s_list]
        pssm_df = pssm_df.iloc[:,:20]

        if pssm_csvFormat_path is not None:
            pssm_df.to_csv(pssm_csvFormat_path)
        else:
            return pssm_df


    # 将原始的pssm转为csv格式提取出来
    def convert_oriPSSM2csvFromat(self, pssmFilePath, csvFilePath=None):
        start = "A   R   N   D   C   Q   E   G   H   I   L   K   M   F   P   S   T   W   Y   V   A   R   N   D   C   Q   E   G   H   I   L   K   M   F   P   S   T   W   Y   V"
        
        endFlagParser = re.compile("K         Lambda")
        spliter = re.compile(r"\s+")
        s_l = spliter.split(start)
        startFlagParser = re.compile(r"\s+".join(s_l))
        selected_amino_acid = "A   R   N   D   C   Q   E   G   H   I   L   K   M   F   P   S   T   W   Y   V"
        saa_l = spliter.split(selected_amino_acid)

        content = []
        readFlag = False

        with open(pssmFilePath,"r") as h:
            for line in h:
                start_r =startFlagParser.search(line)
                end_r = endFlagParser.search(line)
                if end_r:
                    break
                if start_r:
                    readFlag = True
                if readFlag:
                    # content.append(line)
                    l_list = [x for x in spliter.split(line) if x != ""]
                    if l_list == []:
                        continue
                    content.append(l_list)   

        # #去掉1-n行中多余的两列, 并给第1行添加空格，保证数目每行数目相同
        content[0] = [" "]*2 + content[0] + [" "]*2
        # print([len(x) for x in content])

        ct_ary = np.array(content)
        # print(ct_ary.shape)
        m, n = ct_ary.shape
        sub_ary = ct_ary[np.ix_(range(1,m),range(2,n-2))]
        cols = ct_ary[0][2:n-2]
        idx = ct_ary[1:,1]
        # print(cols.shape)
        # print(idx.shape)
        # print(sub_ary.shape)
        df = pd.DataFrame(sub_ary,columns=cols,index=idx)
        # print(df.shape)

        # print(saa_l)
        selected_range = len(saa_l)
        ps_df = df.iloc[:,range(selected_range)]
        if csvFilePath is not None:
            with open(csvFilePath,"w", newline='') as h:
                ps_df.to_csv(h)
        return ps_df
        
    def remove_redundancy_pairwise(self, searchedFilePath,depandentFilePath=None,targetFilePath=None):
        if targetFilePath is None:
            targetFilePath = searchedFilePath

        with open(searchedFilePath,"r") as h:
            sr_df = pd.read_csv(h)
        
        def fetch_no_re(dataframe):
            df_g = dataframe.groupby(["query","template"])
            dg_values = [x[0] for x in df_g.groups.values()]
            no_reDf = dataframe.loc[dg_values]
            return no_reDf
        # # 先去除重复的pairwise
        # srDf_g = sr_df.groupby(["query","template"])
        # sg_values = [x[0] for x in srDf_g.groups.values()]

        # no_reDf = sr_df.loc[sg_values]
        no_reDf = fetch_no_re(sr_df)

        if depandentFilePath is not None:
            # 去除query template属于同一个fold num的情况，这里的fold num同样指交叉验证中的“折”
            with open(depandentFilePath,"rb") as h:
                sn2fN = cPickle.load(h)
            
            # 先取出template为na的部分
            t_df = no_reDf.loc[~pd.isna(no_reDf.loc[:,"template"])]

            # idx = no_reDf.apply(
            #     lambda x: pd.Series(sn2fN[x["query"]] == sn2fN[x["template"]]
            #         ,name="same fold num")
            #     ,axis=1
            # ) # 速度很慢
            qt_df = t_df.loc[:,["query","template"]]
            fN_df = qt_df.apply(lambda d: d.apply(lambda x: sn2fN[x]))
            idx = (fN_df.loc[:,"query"] == fN_df.loc[:, "template"])

            # 对属于同一个fold num 的pairwise的所在项，将除query以外的变成NA
            no_reDf.loc[idx, no_reDf.columns[1:]] = pd.NA

            # 同时要再一次去除冗余的pairwise
            final_df = fetch_no_re(no_reDf)
        else:
            final_df = no_reDf

        with open(targetFilePath,"w",newline='') as h:
            final_df.to_csv(h,index=False)

    # 将csv格式的pssm文件，通过ac, cc 转为(20+380)*LG大小的向量
    def acc_pssm(self, pssm_csvFormat_path, LG):

        with open(pssm_csvFormat_path,"r") as h:
            pssm_df = pd.read_csv(h,index_col=0)

        # 转为frequency
        pssm_df = pssm_df.apply(lambda x: 2**(0.1*x))

        # # 行归一化
        # pssm_df = pssm_df.transform(lambda x: x / x.sum(),axis=1)
        # l1norm = pssm_df.sum(axis=1)
        # pssm_df = pssm_df.div(l1norm,axis=0)

        # # 计算ac
        # ac_l = [ac_calculation(pssm_df.loc[:,x],LG) \
        #     for x in pssm_df.columns]
        # ac_ary = np.hstack(ac_l).astype(np.float32)        

        # # 计算cc
        # column_pair_l = list(combinations(pssm_df.columns,2)) + \
        #     list(combinations(pssm_df.columns[::-1],2))

        # column_pair_l = sorted(column_pair_l)

        # cc_l = [cc_calculation(pssm_df.loc[:,x[0]],pssm_df.loc[:,x[1]],LG) \
        #     for x in column_pair_l]
        # cc_ary = np.hstack(cc_l).astype(np.float32)

        # acc_ary = np.hstack([ac_ary,cc_ary])
        acc_ary = acc_transform(pssm_df.to_numpy(),LG) #快速版本
        return acc_ary
    
    def top_n_gram(self, pssm_csvFormat_path,n=1,L=20):
        pssm_df = pd.read_csv(pssm_csvFormat_path,index_col=0)
        
        # 这里就不需要转成frequency matrix
        # 因为，用于转成frequence matrix的函数是单调递增的

        pssm_mat = pssm_df.to_numpy()

        grams = top_n_gram_trans(pssm_mat,n,L)
        gnorm = np.linalg.norm(grams,1,0)
        grams /= gnorm

        return grams

    def native(self, srch_df, m_df):
        def get_superfamily_df(p_df):
            r_df = m_df.join(p_df.set_index("template"),on="template")
            r_df = r_df.fillna({"prob":0.0,"evalue":1e+6,"score":0.0})
            idx = r_df.groupby("label")["prob"].idxmax().to_numpy(int) # 这里保证了superfamily name是字典序
            sf_df = r_df.loc[idx]
            return sf_df
        
        # 这里保证了sequence name是字典序
        u_df = srch_df.groupby("query").apply(get_superfamily_df)
        u_df = u_df.reset_index(drop=True)
        ucon = tf.constant(
            u_df.loc[:,["prob","evalue","score"]].to_numpy(),
            tf.float32
        )
        return ucon


# 参考这样的指令运行mmseqs
# "mmseqs easy-search {} {} {} {} -e 1e-4 --remove-tmp-files 1 --sort-results 1 --threads 10 --mpi-runner 10"

class MMseqs2(object):
    def __init__(self,queryDBPath=None,targetDBPath=None,alignFilePath=None,tmpPath=None,
        evalue=1e-4,mpi_runner=10,num_iterations=5):
        self.queryDBPath = queryDBPath
        self.targetDBPath = targetDBPath
        self.alignFilePath = alignFilePath
        self.tmpPath = tmpPath
        self.evalue = evalue
        self.mpi_runner = mpi_runner
        self.num_iterations= num_iterations
    
    def createdb(self, set_params):
        def get_createdb_params(iFasFile_list, oSeqDB):
            in_fastaFile = " ".join(iFasFile_list)
            params = "{} {}".format(in_fastaFile,oSeqDB)
            return params
        
        cmd = "mmseqs createdb {}".format(get_createdb_params(**set_params))
        # print(cmd)
        sp.call(cmd,shell=True)
    
    # 注意queryDBPath有可能是多个query
    def easy_search(self, set_params=None):
        def gen_mmseqs_params(queryDBPath,targetDBPath,alignFilePath,tmpPath,
            evalue,mpi_runner,num_iterations):
            params = "mpirun -np {} mmseqs easy-search {} {} {} {} -e {} \
                --num-iterations {} --remove-tmp-files 1 \
                --sort-results 1".format(
                mpi_runner,queryDBPath,targetDBPath,alignFilePath,tmpPath,evalue,
                num_iterations
            )
            return params
        
        if set_params is None:
            set_params = {
                "queryDBPath":self.queryDBPath,
                "targetDBPath": self.targetDBPath,
                "alignFilePath": self.alignFilePath,
                "tmpPath": self.tmpPath,
                "evalue": self.evalue,
                "mpi_runner": self.mpi_runner,
                "num_iterations":self.num_iterations
            }
        params = gen_mmseqs_params(**set_params)

        cmd = params
        # print(cmd)
        err_code = sp.call(cmd,shell=True)
        if err_code != 0 and err_code != 1:
            print("err_code: {}".format(err_code))
    
    # mmseqs是可以把检索结果放到同一个文件里的
    def parsing_searched_result(self, queryList,searchedFilePath, parsingDict=None):
        querySet = set(queryList)
        if parsingDict is None:
            parsingDict = {"query":[],"template":[],
                "prob":[],"evalue":[],"score":[]}
        with open(searchedFilePath,"r") as h:
            for line in h:
                spliter = re.compile(r"\s")
                parsingResult = spliter.split(line.strip())
                query, template,prob = parsingResult[:3]
                evalue, score = parsingResult[-2:]
                parsingDict["query"].append(query)
                parsingDict["template"].append(template)
                parsingDict["prob"].append(float(prob))
                parsingDict["evalue"].append(float(evalue))
                parsingDict["score"].append(float(score))
        existSet = set(parsingDict["query"])
        mis_l = list(querySet.difference(existSet))

        for query in mis_l:
            parsingDict["query"].append(query)
            parsingDict["template"].append(pd.NA)
            parsingDict["prob"].append(pd.NA)
            parsingDict["evalue"].append(pd.NA)
            parsingDict["score"].append(pd.NA)
        
        return parsingDict
    
    def remove_redundancy_pairwise(self, searchedFilePath,depandentFilePath=None,targetFilePath=None):
        if targetFilePath is None:
            targetFilePath = searchedFilePath

        with open(searchedFilePath,"r") as h:
            sr_df = pd.read_csv(h)
        
        def fetch_no_re(dataframe):
            df_g = dataframe.groupby(["query","template"])
            dg_values = [x[0] for x in df_g.groups.values()]
            no_reDf = dataframe.loc[dg_values]
            return no_reDf
        # # 先去除重复的pairwise
        # srDf_g = sr_df.groupby(["query","template"])
        # sg_values = [x[0] for x in srDf_g.groups.values()]

        # no_reDf = sr_df.loc[sg_values]
        no_reDf = fetch_no_re(sr_df)

        if depandentFilePath is not None:
            # 去除query template属于同一个fold num的情况，这里的fold num同样指交叉验证中的“折”
            with open(depandentFilePath,"rb") as h:
                sn2fN = cPickle.load(h)
            
            # 先取出template为na的部分
            t_df = no_reDf.loc[~pd.isna(no_reDf.loc[:,"template"])]

            # idx = no_reDf.apply(
            #     lambda x: pd.Series(sn2fN[x["query"]] == sn2fN[x["template"]]
            #         ,name="same fold num")
            #     ,axis=1
            # ) # 速度很慢
            qt_df = t_df.loc[:,["query","template"]]
            fN_df = qt_df.apply(lambda d: d.apply(lambda x: sn2fN[x]))
            idx = (fN_df.loc[:,"query"] == fN_df.loc[:, "template"])

            # 对属于同一个fold num 的pairwise的所在项，将除query以外的变成NA
            no_reDf.loc[idx, no_reDf.columns[1:]] = pd.NA

            # 同时要再一次去除冗余的pairwise
            final_df = fetch_no_re(no_reDf)
        else:
            final_df = no_reDf

        with open(targetFilePath,"w",newline='') as h:
            final_df.to_csv(h,index=False)

    def native(self, srch_df, m_df):
        def get_superfamily_df(p_df):
            r_df = m_df.join(p_df.set_index("template"),on="template")
            r_df = r_df.fillna({"prob":0.0,"evalue":1e+6,"score":0.0})
            idx = r_df.groupby("label")["prob"].idxmax().to_numpy(int) # 这里保证了superfamily name是字典序
            sf_df = r_df.loc[idx]
            return sf_df
        
        # 这里保证了sequence name是字典序
        u_df = srch_df.groupby("query").apply(get_superfamily_df)
        u_df = u_df.reset_index(drop=True)
        ucon = tf.constant(
            u_df.loc[:,["prob","evalue","score"]].to_numpy(),
            tf.float32
        )
        return ucon

class PsiPred(object):
    def extract_secondary_structure(self,seqPath,targetDir):
        psipred_dataDir = "/home/shaojiangyi/source/Representative_Rank/software/psipred/data"
        weightsPath = os.path.join(psipred_dataDir,"weights.dat")
        weights2Path = os.path.join(psipred_dataDir,"weights.dat2")
        weights3Path = os.path.join(psipred_dataDir,"weights.dat3")
        weights_p2_path = os.path.join(psipred_dataDir,"weights_p2.dat")

        queryName,_ = getFileNameWithoutSurfix(os.path.basename(seqPath))
        
        # 生成mtx文件
        mtxPath = os.path.join(targetDir,"{}.mtx".format(queryName))
        seq2mtx = "seq2mtx {} > {}".format(seqPath,mtxPath)
        r = sp.call(seq2mtx,shell=True)

        assert r == 0

        # 生成ss文件
        ssPath = os.path.join(targetDir,"{}.ss".format(queryName))
        psipred = "psipred {} {} {} {} > {}".format(mtxPath,weightsPath,
            weights2Path,weights3Path,ssPath)
        r = sp.call(psipred,shell=True)

        assert r == 0

        # 删除mtx文件
        rm = "rm {}".format(mtxPath)
        sp.call(rm,shell=True)

        # 生成ss2和horiz文件
        ss2Path = os.path.join(targetDir,"{}.ss2".format(queryName))
        horizPath = os.path.join(targetDir,"{}.horiz".format(queryName))

        psipass2 = "psipass2 {} 1 1.0 1.0 {} {} > {}".format(weights_p2_path,ss2Path,ssPath,horizPath)
        r = sp.call(psipass2,shell=True)

        assert r == 0

        # 删除ss文件
        rm = "rm {}".format(ssPath)
        sp.call(rm,shell=True)
    
    def ss2_to_csvFormat(self, ss2FilePath, csvFilePath=None):
        # 检测到一下行时，就加入到list中
        # 1 F C   0.999  0.000  0.001
        contentParser = re.compile(r"\d+\s+(?P<content>[A-Z]\s+[A-Z](\s+(\d+(\.\d+)?)){3})\s*$")
        # print(contentParser)
        spliter = re.compile(r"\s+")

        content = []
        with open(ss2FilePath, "r") as h:
            for line in h:
                c_r = contentParser.search(line)
                if c_r is not None:
                    l_c = c_r.group("content")
                    l_list = spliter.split(l_c)
                    content.append(l_list)

        # print(content)
        contentAry = np.array(content)

        # probability of folding into each state
        cols = ["state","prob of folding into H","prob of folding into E","prob of folding into C"]
        idx = contentAry[:,0]
        m, n = contentAry.shape

        df = pd.DataFrame(contentAry[:,1:],columns=cols,index=idx)

        if csvFilePath is not None:
            with open(csvFilePath,"w",newline='') as h:
                df.to_csv(h)
        else:
            return df
    
    def represent_ss2_profile(self, ss2_csvFormat_path):
        with open(ss2_csvFormat_path,"r") as h:
            ss2_df = pd.read_csv(h,index_col=0)
        
        m,_ = ss2_df.shape

        # each state / the number of states
        # 计算每个state在所有state中的占比
        m,_ = ss2_df.shape
        states = ["C","E","H"]

        tmp_df = ss2_df.set_index("state")
        def gen(states):
            for s in states:
                try:
                    s_num = tmp_df.loc[s].shape[0]
                    yield s_num
                except:
                    yield 0
        state_num_list = list(gen(states))

        sn_ary = np.array(state_num_list,dtype=np.float32)
        prob_ary = sn_ary / m
        # print(prob_ary)

        # 计算entropy
        log_prob = np.log(prob_ary)
        # print(log_prob)
        p = log_prob * prob_ary
        # print(p)
        mask = np.isnan(p)
        # print(mask)
        p[mask] = 0 # 将nan值替换为0
        # print(p)
        entropy = -np.sum(p)
        # print(entropy)

        # 计算probability series中每列的平均值
        # 就是ss2文件中，那三列数字，计算这三列数据的平均值
        prob_ser_df = ss2_df.iloc[:,1:]
        # prob_ser_df
        pSer_ary = prob_ser_df.to_numpy().astype(np.float32)
        # pSer_ary
        ave_ary = np.average(pSer_ary,0)
        # print(ave_ary)

        # # 计算probability series的acc
        # # 把这三列看成矩阵，把ac和cc用于该矩阵
        # # 行归一化
        # prob_ser_df = prob_ser_df.transform(lambda x: x / x.sum(),axis=1)
        # l1norm = prob_ser_df.sum(axis=1)
        # prob_ser_df = prob_ser_df.div(l1norm,axis=0)

        # def ac_trans(x):
        #     LG = 9
        #     return ac_calculation(x,LG)

        # def cc_trans(x1,x2):
        #     LG = 9
        #     return cc_calculation(x1,x2,LG)

        # ac_l = [ac_trans(prob_ser_df.loc[:,x]) \
        #     for x in prob_ser_df.columns]
        # ac_ary = np.hstack(ac_l).astype(np.float32)
        # # print(ac_ary)

        # comb_pair_l = list(combinations(prob_ser_df.columns,2)) +\
        #     list(combinations(prob_ser_df.columns[::-1],2))

        # comb_pair_l = sorted(comb_pair_l)

        # cc_l = [cc_trans(prob_ser_df.loc[:,x[0]],prob_ser_df.loc[:,x[1]]) \
        #     for x in comb_pair_l]
        # cc_ary = np.hstack(cc_l).astype(np.float32)
        # # print(cc_ary)

        # acc_ary = np.hstack([ac_ary,cc_ary])
        # # print(acc_ary)

        # 快速版本
        acc_ary = acc_transform(prob_ser_df.to_numpy(),9)

        # 对该矩阵使用bi-gram
        def bi_gram_calculation(mat):
            r, _ = mat.shape
            # def gen(mat,L, c):
            #     for m in range(c):
            #         for n in range(c):
            #             # p = sum([mat[i][m]*mat[i+1][n] \
            #             #     for i in range(L-1)])
            #             p = mat[:L-1,m] @ mat [1:, n] #速度更快
            #             yield p
            # bi_gram = list(gen(mat,r, c))

            # 本质上这里就是矩阵乘法
            bi_gram = np.hstack(mat[:r-1].T @ mat[1:])
            return bi_gram
        bg_ary = bi_gram_calculation(pSer_ary)
        # print(bg_ary)

        # 对该矩阵使用tri-gram
        # tri-gram不方便化成矩阵乘法，但可以通过向量乘法来实现
        def tri_gram_calculation(mat):
            r, c = mat.shape
            def gen(mat,L, c):
                for m in range(c):
                    for n in range(c):
                        for x in range(c):
                            # p = sum([mat[i][m]*mat[i+1][n]*mat[i+2][x] \
                            #     for i in range(L-2)]) # strategy 1
                            p = mat[:L-2,m] * mat[1:L-1,n] * mat[2:,x] # strategy 2
                            yield p
            # tri_gram = list(gen(mat,r,c)) # strategy 1
            tri_gram = np.array(list(gen(mat,r, c))).sum(1) # strategy 2
            # strategy 2的速度更快
            return tri_gram
        tg_ary = tri_gram_calculation(pSer_ary)
        # print(tg_ary)

        # 将所有数据整合
        ssv_ary = np.hstack([prob_ary,entropy,ave_ary,acc_ary,bg_ary,tg_ary]).astype(np.float32)
        # print(ssv_ary.shape)
        # print(ssv_ary)
        return ssv_ary



# mpirun -np 10 hhblits_mpi -i {} -d {} -o {} -n 2 -cpu 1 -maxfilt 500000 -diff inf -id 99 -cov 60
class HHblits(object):
    def fas2ffFile(self, fastaFilePath, ffFilePath):
        ffdataPath = "{}.ffdata".format(ffFilePath)
        ffindexPath = "{}.ffindex".format(ffFilePath)
        ffindex_from_fasta = "ffindex_from_fasta -s {} {} {}".format(ffdataPath,ffindexPath,fastaFilePath)
        # print(ffindex_from_fasta)
        sp.call(ffindex_from_fasta,shell=True)

    def search(self,set_params):
        def get_cmd(input,db,output,iteration,cpu_core,maxfilt,diff,id,cov,np):
            cmd = "mpirun -np {} \
                hhblits_mpi -i {} -d {} -o {} -p 0\
                    -n {} -cpu {} -maxfilt {} -diff {} -id {} -cov {}".format(
                    np,input,db,output,iteration,cpu_core,maxfilt,diff,id,cov
                )
            return cmd
        cmd = get_cmd(**set_params)
        # print(cmd)
        sp.call(cmd,shell=True)
    
    def search_by_hmm(self, set_params):
        def get_cmd(input,db,output,iteration,evalue,cpu_core,maxfilt,diff,id,cov,np):
            cmd = "mpirun -np {} \
                hhblits_mpi -i {} -d {} -o {} -e {}\
                    -n {} -cpu {} -maxfilt {} -diff {} -id {} -cov {}".format(
                    np,input,db,output,evalue,iteration,cpu_core,maxfilt,diff,id,cov
                )
            return cmd
        cmd = get_cmd(**set_params)
        # print(cmd)
        sp.call(cmd,shell=True)        
    
    def hhsearch(self, set_params):
        def get_cmd(input,db,output,cpu_core,diff,id,cov,np):
            cmd = "mpirun -np {} \
                hhsearch_mpi -i {} -d {} -o {}\
                 -cpu {} -diff {} -id {} -cov {}".format(
                    np,input,db,output,cpu_core,diff,id,cov
                )
            return cmd
        cmd = get_cmd(**set_params)
        # print(cmd)
        sp.call(cmd,shell=True)
    
    def hhsearch_by_hmm(self, set_params):
        def get_cmd(input,db,output,evalue,cpu_core,diff,id,cov,np):
            cmd = "mpirun -np {} \
                hhsearch_mpi -i {} -d {} -o {} -e {}\
                    -cpu {} -diff {} -id {} -cov {}".format(
                    np,input,db,output,evalue,cpu_core,diff,id,cov
                )
            return cmd
        cmd = get_cmd(**set_params)
        # print(cmd)
        sp.call(cmd,shell=True)        
    
    def extract(self, kind):
        kind = "-" + kind
        cmd = "mpirun -np {} \
            hhblits_mpi -i {} -d {} " + kind + " {} -n {} -cpu {} -maxfilt {} -diff {} -id {} -cov {}"
        return cmd

    def extract_a3m(self, set_params):
        def get_cmd(input,db,output,iteration,cpu_core,maxfilt,diff,id,cov,np):
            form = self.extract("oa3m")
            cmd = form.format(np,input,db,output,iteration,cpu_core,maxfilt,diff,id,cov)
            return cmd
        cmd = get_cmd(**set_params)
        # print(cmd)
        sp.call(cmd,shell=True)
    
    def a3m2hmm(self, a3mFilePath, hmmFilePath,np=10):
        # mpirun -np <number_threads> ffindex_apply_mpi <db>_a3m.ff{data,index} \\
        # -i <db>_hhm.ffindex -d <db>_hhm.ffdata -- hhmake -i stdin -o stdout -v 0

        a3mDataPath = "{}.ffdata".format(a3mFilePath)
        a3mIndexPath = "{}.ffindex".format(a3mFilePath)
        hhmDataPath = "{}.ffdata".format(hmmFilePath)
        hhmIndexPath = "{}.ffindex".format(hmmFilePath)
        mpirun = "mpirun -np {} ffindex_apply_mpi {} {} -d {} -i {} -- hhmake -i stdin -o stdout -v 0"

        cmd = mpirun.format(
                np,
                a3mDataPath,a3mIndexPath,hhmDataPath,hhmIndexPath
            )
        sp.call(cmd,shell=True)
        # print(cmd) # 用于查看命令行是否正确
    
    def a3m2cs219(self, a3mFildPath, cs219FilePath,np = 10):
        # mpirun -np <number_threads> cstranslate_mpi -f -x 0.3 -c 4 -I a3m -i <db>_a3m -o <db>_cs219

        mpirun = "mpirun -np {} cstranslate_mpi -f -x 0.3 -c 4 -I a3m -i {} -o {}"

        cmd = mpirun.format(np,a3mFildPath,cs219FilePath)
        sp.call(cmd,shell=True)
        # print(cmd) # 用于查看命令行是否正确
    
    def parsing_searched_result(self,searchedFilePath):
        def genValue(handle):
            startLine = "No Hit                             Prob E-value P-value  Score    SS Cols Query HMM  Template HMM"
            spliter = re.compile(r"\s+")
            startFlagParser = re.compile(r"\s+".join(spliter.split(startLine)))
            # print(startFlagParser)
            newLineParser = re.compile(r"^\n")
            # Query         d2vxfa1 b.38.1.5
            queryParser = re.compile(r"Query\s+(?P<name>[\w_\.]+)\s+[a-z]\.(\d+\.?){3}\s+")
            # 1 d2vxfa1 b.38.1.5
            # lineContentParser = re.compile(r"\s+(?P<content>[0-9]+\s+[\w_\.]+\s+[a-z]\.[0-9]+\.[0-9]+\.[0-9]+)\s+")
            lineContentParser = re.compile(r"\s+[0-9]+\s+(?P<content>[\w_\.]+)\s+[a-z]\.(\d+\.?){3}\s+") #简洁一点
            # contentFetchingParser = re.compile(r"\s+(?P<content>(\S+\s*){16}")) # 因为hhr文件中类似'B core S'的使得不太好确定一行中空格的分布规律，所以放弃这种parser
            # 之所以不直接取下一行，还是因为上述问题，导致不方便解析，因此，直接在这里取出value
            # 新的方法是找到value的位置，格式为‘94.2 0.00045 1.1E-07   38.1   0.0   77    1-78     10-87  (90)’
            # 同时放弃找寻最后一个(90)，因为cols总共就8个
            # valueStartParser = "\(?\d+((\.\d+(E\-\d+)?)|(-\d+))?\)?" #可以找到所有的，但没必要
            # valueStartParser = re.compile(r"(\d+(\.\d+(E\-\d+){0,1}|-\d+)?\s*){7}\d+(\.\d+(E\-\d+){0,1}|-\d+)?") # 忽略了1E-90这种情况
            # valueStartParser = re.compile(r"(?P<value>((\d+(\.\d+)?)(E\-\d+|-\d+){0,1}\s*){7}(\d+(\.\d+)?)(E\-\d+|-\d+){0,1})\s*(\(\d+\)\s*)$") # strategy 1 忽略了1E+09这种情况
            valueStartParser = re.compile(r"(?P<value>((\d+(\.\d+)?)(E[+\-]\d+|\-\d+){0,1}\s*){7}(\d+(\.\d+)?)(E[+\-]\d+|\-\d+){0,1})\s*(\(\d+\)\s*)$") # strategy 1


            # 如果担心太长容易出错,就直接使用spliter将line分割,然后,倒着数9个,并去掉最后一个即可,参考strategy 2
            # strategy 2是错误的，有时候最后两列之间会没有空格，导致直接用空格来分会出错

            newLineParser = re.compile("^\n")

            readFlag = False

            for line in handle:
                if not readFlag:
                    q_r = queryParser.search(line)
                    sF_r = startFlagParser.search(line)
                    if q_r is not None:
                        query = q_r.group("name")
                
                    if sF_r is not None:
                        readFlag = True
                if readFlag:
                    nL_r = newLineParser.search(line)
                    if nL_r is not None:
                        readFlag = False
                    else:
                        lC_r = lineContentParser.search(line)
                        
                        if lC_r is not None:
                            l_list = [query]
                            l_list.append(lC_r.group("content"))
                            
                            # strategy 1
                            val_r = valueStartParser.search(line[lC_r.end():])
                            v_list = spliter.split(val_r.group("value"))[-8:-4]
                            
                            # # strategy 2
                            # v_list = spliter.split(line[lC_r.end():-1])[-9:-5]
                            
                            l_list += v_list
                            yield l_list
            return
    
        with open(searchedFilePath,"r") as h:
            buffer = list(genValue(h))
        
        return buffer

    # 因为使用hhblits查询的时候是对着一个数据集查询的因此，会出现重复的情况
    # depandentFilePath是一个字典型的pickle文件，格式为<sequence name: fold num>
    def remove_redundancy_pairwise(self, searchedFilePath,depandentFilePath=None,targetFilePath=None):
        if targetFilePath is None:
            targetFilePath = searchedFilePath

        with open(searchedFilePath,"r") as h:
            sr_df = pd.read_csv(h)
        
        def fetch_no_re(dataframe):
            df_g = dataframe.groupby(["query","template"])
            dg_values = [x[0] for x in df_g.groups.values()]
            no_reDf = dataframe.loc[dg_values]
            return no_reDf
        # # 先去除重复的pairwise
        # srDf_g = sr_df.groupby(["query","template"])
        # sg_values = [x[0] for x in srDf_g.groups.values()]

        # no_reDf = sr_df.loc[sg_values]
        no_reDf = fetch_no_re(sr_df)

        if depandentFilePath is not None:
            # 去除query template属于同一个fold num的情况，这里的fold num同样指交叉验证中的“折”
            with open(depandentFilePath,"rb") as h:
                sn2fN = cPickle.load(h)
            
            # 先取出template为na的部分
            t_df = no_reDf.loc[~pd.isna(no_reDf.loc[:,"template"])]

            # idx = no_reDf.apply(
            #     lambda x: pd.Series(sn2fN[x["query"]] == sn2fN[x["template"]]
            #         ,name="same fold num")
            #     ,axis=1
            # ) # 速度很慢
            qt_df = t_df.loc[:,["query","template"]]
            fN_df = qt_df.apply(lambda d: d.apply(lambda x: sn2fN[x]))
            idx = (fN_df.loc[:,"query"] == fN_df.loc[:, "template"])

            # 对属于同一个fold num 的pairwise的所在项，将除query以外的变成NA
            no_reDf.loc[idx, no_reDf.columns[1:]] = pd.NA

            # 同时要再一次去除冗余的pairwise
            final_df = fetch_no_re(no_reDf)
        else:
            final_df = no_reDf

        with open(targetFilePath,"w",newline='') as h:
            final_df.to_csv(h,index=False)
    
    def hmm2csvFormat(self, hmmFilePath, csvFilePath):
        hmmContent = []
        spliter = re.compile(r"\s+")
        startLine = "HMM    A	C	D	E	F	G	H	I	K	L	M	N	P	Q	R	S	T	V	W	Y"
        startFlagParser = re.compile(r"\s+".join(spliter.split(startLine)))
        endFlagParser = re.compile(r"^//")
        ignoreFlagParser = re.compile(r"^\n")

        readFlag = False

        with open(hmmFilePath,"r") as h:
            for line in h:
                if not readFlag:
                    sF_r = startFlagParser.search(line)
                    if sF_r is not None:
                        readFlag = True
                
                if readFlag:
                    eF_r = endFlagParser.search(line)
                    if eF_r is not None:
                        break
                    else:
                        iF_r = ignoreFlagParser.search(line)
                        if iF_r is not None:
                            continue
                        else:
                            hmmContent.append(line)

        newContent = []
        # 忽略前三行，因为，在我们选取的这些内容中，前三行是和dataframe的key有关的
        for i in range(3,len(hmmContent)-1,2):
            line = hmmContent[i] + hmmContent[i+1]
            l_list = [x for x in spliter.split(line) if x != ""]
            newContent.append(l_list)

        colsLine = "index1 A	C	D	E	F	G	H	I	K	L	M	N	P	Q	R	S	T	V	W	Y index2\
             M->M	M->I	M->D	I->M	I->I	D->M	D->D	Neff	Neff_I	Neff_D" #第一位多一个空格
        cols = spliter.split(colsLine)

        contentAry = np.array(newContent)
        idx = contentAry[:,0]
        contentAry = contentAry[:,1:]
        df = pd.DataFrame(contentAry,columns=cols,index=idx)
        df = df.drop(["index1","index2"],axis=1)

        with open(csvFilePath,"w",newline='') as h:
            df.to_csv(h)
    
    def acc_hmm(self,hmm_csvFormat_path,LG=5):
        with open(hmm_csvFormat_path,"r") as h:
            hmm_df = pd.read_csv(h,index_col=0)

        h_df = hmm_df.iloc[:,:20]
        # print(h_df.shape)
        # print(h_df)
        h_df.replace(['*'],np.inf,inplace=True)
        # print(h_df)
        h_df = h_df.astype(np.float32)
        h_df = 2**(-1e-3*h_df)
        # print(h_df)

        # # # 行归一化
        # h_df = h_df.transform(lambda x: x / x.sum(),axis=1)
        # l1norm = h_df.sum(axis=1)
        # h_df = h_df.div(l1norm,axis=0)

        # def ac_trans(x):
        #     LG = 5
        #     return ac_calculation(x,LG)

        # def cc_trans(x1,x2):
        #     LG = 5
        #     return cc_calculation(x1,x2,LG)

        # comb_pair_l = list(combinations(h_df.columns,2)) +\
        #     list(combinations(h_df.columns[::-1],2))

        # comb_pair_l = sorted(comb_pair_l)

        # ac_l = [ac_trans(h_df.loc[:,x]) \
        #     for x in h_df.columns]
        # # print(ac_l)
        # ac_ary = np.hstack(ac_l).astype(np.float32)
        # # print(ac_ary)

        # cc_l = [cc_trans(h_df.loc[:,x[0]],h_df.loc[:,x[1]]) \
        #     for x in comb_pair_l]
        # cc_ary = np.hstack(cc_l).astype(np.float32)
        # # print(cc_ary)
        # rep_ary = np.hstack([ac_ary,cc_ary])
        # # print(rep_ary.shape)
        # # print(rep_ary)
        rep_ary = acc_transform(h_df.to_numpy(),LG)
        return rep_ary
    
    # def bigram_hmm(self, hmm_csvFormat_path):
    def top_n_gram(self, hmm_csvFormat_path,n=1,L=20):
        hmm_df = pd.read_csv(hmm_csvFormat_path,index_col=0)
        h_df = hmm_df.iloc[:,:20]

        h_df.replace(['*'],np.inf,inplace=True)
        h_df = h_df.astype(np.float32)
        h_df = 2**(-1e-3*h_df)
        # 这里就不需要转成frequency matrix
        # 因为，用于转成frequence matrix的函数是单调递增的
        h_mat = h_df.to_numpy()
        grams = top_n_gram_trans(h_mat,n,L)
        gnorm = np.linalg.norm(grams,1,0)
        grams /= gnorm
        
        return grams
    
    # 提取native特征
    # 原本的排序结果是pairwise类型的，现在转为seq2cluster
    # 方法是，每个superfamily选取一个最好的结果，如果某个superfamily没有对应的序列在反馈列表中则用0表示
    # 最终得到n_class*3的矩阵
    # 在外部把所有结果合成n_sample*n_class*3的矩阵
    # 这个'3'表示，prob, evalue, score这3种结果
    # srch_df为包含query, template, prob, evalue, score的dataframe
    # m_df表示template, row num, sf, label
    # srch_df和m_df符合5折交叉验证的划分
    def native(self, srch_df, m_df):
        def get_superfamily_df(p_df):
            r_df = m_df.join(p_df.set_index("template"),on="template")
            r_df = r_df.fillna({"prob":0.0,"evalue":1e+6,"score":0.0})
            idx = r_df.groupby("label")["prob"].idxmax().to_numpy(int) # 这里保证了superfamily name是字典序
            sf_df = r_df.loc[idx]
            return sf_df
        
        # 这里保证了sequence name是字典序
        hg = srch_df.groupby("query")
        query_count = len(hg) # 得到query的个数，方便进行reshape
        u_df = hg.apply(get_superfamily_df)
        u_df = u_df.reset_index(drop=True) # 因为使用reset_index后query会overlap，因此，drop=True
        ucon = tf.constant(
            u_df.loc[:,["prob","evalue","score"]].to_numpy(),
            tf.float32
        )
        ucon = tf.reshape(ucon, (query_count,-1,ucon.shape[-1]))
        return ucon


class PseAAC(object):
    # 对fasta文件中每条序列进行处理
    def representation(self, fasPath):
        from Bio.SeqIO import parse
        from functools import reduce
        

        csvPath = "feature/physicochemical/ori/amphiphilic_pseudo-amino_acid_components.csv"

        pse_df = pd.read_csv(csvPath)

        h1_std = pse_df.loc[:,"hydrophobicity"].std()
        h1_mean = pse_df.loc[:,"hydrophobicity"].mean()
        h2_std = pse_df.loc[:,"hydrophilicity"].std()
        h2_mean = pse_df.loc[:,"hydrophilicity"].mean()
        # print(h1_std,h2_std)
        # print(h1_mean,h2_mean)
        # p相当于原文里的lambda
        pse_df.set_index("amino acid",inplace=True)
        def get_sequenceOrder_effects(protein_seq,p):
            # 计算the hydrophobicity (h1) and hydrophilicity (h2) correlation
            # def h1_correlation(r_i,r_j):
            #     df = pse_df.set_index("amino acid")
            #     h1_i = (df.loc[r_i,"hydrophobicity"] - h1_mean) / h1_std
            #     h1_j = (df.loc[r_j,"hydrophobicity"] - h1_mean) / h1_std
            #     return h1_i*h1_j
            
            # def h2_correlation(r_i,r_j):
            #     df = pse_df.set_index("amino acid")
            #     h2_i = (df.loc[r_i,"hydrophilicity"] - h2_mean) / h2_std
            #     h2_j = (df.loc[r_j,"hydrophilicity"] - h2_mean) / h2_std
            #     return h2_i*h2_j
            
            L = len(protein_seq)
            
            h1_ary = (pse_df.loc[list(protein_seq),"hydrophobicity"].to_numpy() - h1_mean) / h1_std
            h2_ary = (pse_df.loc[list(protein_seq),"hydrophilicity"].to_numpy() - h2_mean) / h2_std
            rep_l = []
            for l in range(1,p+1):
                # t1 = 1 / (L-l) * sum([h1_correlation(protein_seq[i],protein_seq[i+l]) \
                #     for i in range(L-l)])
                # t2 = 1 / (L-l) * sum([h2_correlation(protein_seq[i],protein_seq[i+l]) \
                #     for i in range(L-l)])
                t1 = 1 / (L-l) * (h1_ary[:L-l] @ h1_ary[l:])
                t2 = 1 / (L-l) * (h2_ary[:L-l] @ h2_ary[l:])

                rep_l += [t1, t2]

            amino_acid_list = "ACDEFGHIKLMNPQRSTVWY"
            amino_acid_pair_list = [(a, 0) for a in amino_acid_list]
            ap_dict = dict(amino_acid_pair_list)
            # amino_acid_pair_list
            for x in protein_seq:
                ap_dict[x] += 1

            freq_ary = np.array(list(ap_dict.values()))
            # print(freq_ary.shape)
            freq_ary = freq_ary / L
            # freq_ary

            w = 0.5
            rep_ary = np.array(rep_l)
            pse_aac_ary = np.hstack([freq_ary, w*rep_ary])
            norm1 = np.linalg.norm(pse_aac_ary,1,0)
            # print(norm1,np.sum(np.abs(pse_aac_ary)))

            pse_aac_ary = pse_aac_ary / norm1
            # pse_aac_ary
            return pse_aac_ary

        def gen(handle):
            sub_parser = re.compile(r"[BZX\*\s]")
            for record in parse(handle,"fasta"):
                seq = record.seq
                seq = sub_parser.sub('',str(seq).upper())
                # print(seq)
                p = 5

                pse_aac_ary = get_sequenceOrder_effects(seq,p)
                yield pse_aac_ary
        
        with open(fasPath,"r") as h:
            rep_mat = np.vstack(list(gen(h)))
        
        return rep_mat
            
