from math import ceil
from src.utils.tools import Assembly
from Bio.SeqIO import parse, write
import lightgbm as lgb
import numpy as np
import pandas as pd
import time
import os
import subprocess as sp
import tensorflow as tf
from src.model.learningtorank import LambdaMART
from src.evaluation.cross_validation import Run
from sklearn.metrics import roc_auc_score
from sklearn.preprocessing import label_binarize
import mpi4py.MPI as MPI
from functools import reduce

toola = Assembly()
run_lambdamart = LambdaMART()
run_cv = Run()

#建立一个training_ltr的模板
# 其他函数实现一个装饰器即可
#%%
def gen_training_ltr_template(process_func):
    def gen_training_ltr_function(i, **kArgs):
        fn_col = str(i)
        params = kArgs["params"]
        train_args = kArgs["train_args"]
        template_df = kArgs["template_df"]
        vtemplate_df = kArgs["vtemplate_df"]
        cv_df = kArgs["cv_df"]
        k_df = kArgs["k_df"]
        kv_df = kArgs["kv_df"]
        representation_dict = kArgs["representation_dict"]
        vrepresentation_dict = kArgs["vrepresentation_dict"]
        norm_flag = kArgs["norm_flag"]
        use_eval_flag = kArgs["use_eval_flag"]
        eval_norm = kArgs["eval_norm"]
        fn_col = str(i)
        t_i_df = template_df.loc[~template_df.loc[:,"template"].isin(cv_df.loc[:,fn_col])]
        kv_i_df = k_df.loc[~k_df.loc[:,"template"].isin(cv_df.loc[:,fn_col])]
        kv_i_df = kv_i_df.sort_values(by="label")
        representation_l = representation_dict[i]
        vrepresentation_l = vrepresentation_dict[i]
        # 根据sequence identity调整训练集，验证集
        con = run_lambdamart.get_training_set(representation_l)
        # k_df不按交叉验证的划分结果去处理，是因为这里不仅仅处理训练集和验证集，还会处理测试集
        label = run_lambdamart.get_label(kv_df,kv_i_df, t_i_df,True)

        # 补充process_func需要的参数
        kArgs["t_i_df"] = t_i_df
        con, label = process_func(con,label, **kArgs)

        # 目前con变成了selected_samples * n_class * k大小
        group = con.shape[1]*np.ones(con.shape[0])
        # 构造dataset前对con和vcon进行reshape
        con = tf.reshape(con,(-1,con.shape[-1]))
        # label和vlabel目前也是n_sample*n_class大小，需要reshape
        label = tf.reshape(label,(-1,))
        train_set = lgb.Dataset(con.numpy(),label.numpy(),group=group)
        train_args["train_set"] = train_set
        if use_eval_flag:
            vcon = run_lambdamart.get_training_set(vrepresentation_l)
            vlabel = run_lambdamart.get_label(kv_df, kv_i_df, vtemplate_df,True)
            # 归一化
            if eval_norm:
                vcon = run_lambdamart.normalization(vcon,2)
            # 同理
            vgroup = vcon.shape[1]*np.ones(vcon.shape[0])
            vlabel = tf.reshape(vlabel,(-1,))
            vcon = tf.reshape(vcon,(-1,vcon.shape[-1]))
            val_set = lgb.Dataset(vcon.numpy(),vlabel.numpy(),group=vgroup)
            train_args["valid_sets"] = [val_set]
        return {"params":params,"train_args":train_args}    
        
    return gen_training_ltr_function

def gen_test_ltr_template(process_func):
    def gen_test_ltr_func(i, **kArgs):
        fn_col = str(i)

        cv_df = kArgs["cv_df"]
        k_df = kArgs["k_df"]
        eval_norm = kArgs["eval_norm"]
        save_booster_l = kArgs["save_booster_l"]
        eval_norm = kArgs["eval_norm"]
        load_path = save_booster_l[i]
        booster = lgb.Booster(model_file=load_path)
        kv_i_df = k_df.loc[~k_df.loc[:,"template"].isin(cv_df.loc[:,fn_col])]
        kv_i_df = kv_i_df.sort_values(by="label")

        # 补齐需要的参数
        kArgs["kv_i_df"] = kv_i_df
        con, label = process_func(i, **kArgs)

        if eval_norm:
            con = run_lambdamart.normalization(con,2)
        return (con, label, booster)
    return gen_test_ltr_func

def test_ltr_template(gclb_func):
    def test_ltr_func(n, **kArgs):
        save_flag = kArgs["save_flag"] # 保存预测结果
        save_output_flag = kArgs["save_output_flag"] # 保存评估方法roc1, roc50的结果
        save_l = kArgs["save_l"]
        eval_print_path = kArgs["eval_print_path"]

        output_file_path = kArgs["output_file_path"] # 保存评估方法roc1, roc50结果的路径
        def iter_func(*args):
            con, label, booster = args
            X = con.numpy()
            y_test = label.numpy()
            gen_pred_test = map(lambda j: [booster.predict(X[j], raw_score=True),y_test[j]], range(X.shape[0]))
            return gen_pred_test

        def single_eval(*args):
            y_pred, y_test = args
            indices = tf.argsort(y_pred,axis=0,direction="DESCENDING").numpy()
            yt = y_test[indices]
            score1 = run_cv.cal_roc_mu_at(1,yt)
            score50 = run_cv.cal_roc_mu_at(50,yt)
            try:
                ascore = roc_auc_score(y_test,y_pred)
            except:
                ascore = 0.0
            return [score1,score50,ascore]
        
        def iter_save(*args):
            ary,save_path = args
            np.save(save_path,ary)
        
        gpt_list = [iter_func(*gclb_func(i,**kArgs)) for i in range(n)]
        # n_fold * each_fold_seq * 2 * each_fold_class
        # 调试一下
        # debug = [list(gpt) for gpt in gpt_list]
        # 注意每一折的class数目是不同的
        # 但同一折内的class数目是相同的
        pt_nested_list = [list(gpt) for gpt in gpt_list]
        # n_samples * 2 * each_fold_class
        pt_list = reduce(lambda x,y:x+y,pt_nested_list)
        roc_ary = np.array(list(map(lambda p: single_eval(*p), pt_list))) # 评估结果
        score = np.mean(roc_ary,0)

        pt_ten_list = list(map(tf.constant,pt_nested_list))
        # 变成n_fold * each_fold_seq * each_fold_class
        pt_ten_list = list(map(lambda t: tf.transpose(t,[1,0,2])[0],pt_ten_list))

        if save_flag:
            list(map(lambda p:iter_save(*p), zip(pt_ten_list,save_l)))

        # 打印最终结果
        if save_output_flag:
            np.savetxt(output_file_path,roc_ary,"%.16f")
        
        with open(eval_print_path,"a") as h:
            line = "{},{},{}\n".format(*score)
            h.write(line)

        print(score)
    return test_ltr_func

def training_ltr_template(gen_lgb_config_func):
    def training_ltr_func(n, **kArgs):
        save_booster_l = kArgs["save_booster_l"]
        def call_ltr_by_config(params,train_args):
            return run_lambdamart.train(params,**train_args)
        
        def iter_func(i):
            config = gen_lgb_config_func(i,**kArgs)
            return call_ltr_by_config(**config)
        
        def iter_save_model(*args):
            booster, save_path = args
            booster.save_model(save_path)

        booster_list = list(map(iter_func,range(n)))
        gen_result = map(
            lambda p:iter_save_model(*p),
            zip(booster_list,save_booster_l)
        )
        # 执行iterator
        list(gen_result)
    return training_ltr_func

def float2fs(f):
    s = str(round(f,3))
    import re
    srepl = re.compile("\.")
    fs = srepl.sub("_",s)
    return fs

def use_cdhit(fas_path,clu_path,sequence_identity,word_size,threads=10):
    # 在use_cdhit里面会根据sequence_identity大小自动调用不同的程序
    toola.use_cdhit(fas_path,clu_path,sequence_identity,word_size,threads)


def sidentity2wordsize(siden):
    # 当sequence_identity小于0.4时，word_size这个参数不会被使用到，因此没事
    assert siden <= 1.0
    if siden < 0.5:
        word_size = 2
    elif siden < 0.6:
        word_size = 3
    elif siden < 0.7:
        word_size = 4
    else:
        word_size = 5
    return word_size
    
# 设置不同的sequence identity
# 讨论不同情况下，feature importance与roc1, roc50的关系
def test_sequence_identity():
    fas_path = "dataset/scop_scope/ori/merge.fa"
    clu_path = "dataset/scop_scope/ori/merge_cdhit_{}.fa"

    # si0, si1, step = 0.4, 1.0, 0.05
    si0, si1, step = 0.25, 0.4, 0.05 # 更低sequence identity的情况
    config = {"si":(si0,si1,step)}


    word_size_gen = map(sidentity2wordsize, np.arange(si0,si1,step))
    # for test_sequence_identity, word_size in zip(np.arange(0.4,0.9,0.05),word_size_gen):
    gen_result = map(
        lambda p: use_cdhit(fas_path, clu_path.format(float2fs(p[0])), p[0],p[1]),
        zip(np.arange(si0,si1,step),word_size_gen)
    )
    list(gen_result)

    return config

def single_sequence_identity(single_seqiden_path,sequence_identity_path):
    # sequence_identity = 0.95
    sequence_identity = 0.35
    word_size = 5
    fas_path = "dataset/scop_scope/ori/merge.fa"
    clu_path = "dataset/scop_scope/ori/merge_cdhit_{}.fa".format(float2fs(sequence_identity))
    config = dict(
        seqiden = sequence_identity,
        word_size = word_size,
        fas_path = fas_path,
        clu_path = clu_path,
        sgle_path = single_seqiden_path,
        diff_path = sequence_identity_path
    )
    
    use_cdhit(fas_path, clu_path, sequence_identity, word_size)
    return config

def fas2series(fas_path):
    with open(fas_path,"r") as h:
        seq_list = [r.name for r in parse(h,"fasta")]
    
    ser = pd.Series(seq_list)
    return ser

# 制作这样一个dataframe
# 0.4表示是否属于sequence identity=0.4这个数据集
# template, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9
def sequence2benchmark(diff_path,**kArgs):
    fas_path = "dataset/scop_scope/ori/merge.fa"
    clu_path = "dataset/scop_scope/ori/merge_cdhit_{}.fa"

    with open(fas_path,"r") as h:
        seq_list= [r.name for r in parse(h,"fasta")]
    
    ser0 = pd.Series(seq_list)

    def isin_fas(fas_path):
        with open(fas_path,"r") as h:
            seq_list = [r.name for r in parse(h,"fasta")]
        sub = pd.Series(seq_list)
        indices = ser0.isin(sub)
        return indices
    # si0, si1, step = 0.4, 0.95, 0.05
    si0,si1,step = kArgs["si"]
    gen_fas_list = map(lambda s: clu_path.format(float2fs(s)),np.arange(si0,si1,step))
    indices_list = list(map(isin_fas,gen_fas_list))

    n = (si1-si0) / step
    s2bch = pd.concat([ser0]+indices_list,axis=1)
    s2bch = s2bch.rename(
        columns=dict(
            zip(
                list(range(1,ceil(n)+1)),
                [round(x,2) for x in np.arange(si0,si1,step)]
            )
        )
    )
    s2bch = s2bch.rename(columns={0:"template"})
    s2bch.to_csv(diff_path,index=False)

# 把0.95那一列也添加进去
def add_sequence_identity(df, sequence_identity, fas_path):
    with open(fas_path,"r") as h:
        seq_list = [r.name for r in parse(h,"fasta")]
    
    sub = pd.Series(seq_list)
    indices = df.iloc[:,0].isin(sub)
    df.loc[:,sequence_identity] = indices
    return df

def single_seq2bench(df, sequence_identity, fas_path):
    with open(fas_path,"r") as h:
        seq_list = [r.name for r in parse(h,"fasta")]
    
    sub = pd.Series(seq_list)
    indices = df.iloc[:,0].isin(sub)

    df.loc[:,sequence_identity] = indices
    sdf = df.iloc[:,[0,-1]]
    return sdf

def call_single_seq2bench(**kArgs):
    diff_path = kArgs["diff_path"]
    fas_path = kArgs["clu_path"]
    save_path = kArgs["sgle_path"]
    seqiden = kArgs["seqiden"]
    diffdf = pd.read_csv(diff_path)
    single_df = single_seq2bench(diffdf,seqiden,fas_path)
    single_df.to_csv(save_path,index=False)

def call_add_sequence_identity(**kArgs):
    diff_path = kArgs["diff_path"]
    fas_path = kArgs["clu_path"]
    save_path = diff_path # 用新的结果替换原来的结果
    seqiden = kArgs["seqiden"]
    diffdf = pd.read_csv(diff_path)
    nddf = add_sequence_identity(diffdf,seqiden,fas_path)
    nddf.to_csv(save_path,index=False)


#-------------------------------------------------------------------------------------------------------------------------------
# 根据sequence identity，重新训练模型并进行独立测试集上的相关测试
#-------------------------------------------------------------------------------------------------------------------------------
# 实验如下
def init_training_config(boosting_type, use_eval_flag, use_test_flag, norm_flag, eval_norm,save_booster_l):
    # learning_rate = 0.07 # 在使用psiblastexb情况下
    # 改变特征后发现0.059的学习率使得训练集的拟合度不及原来
    # 但是，感觉继续改变LinearSVM(HMM)的话效果又会恢复
    # 因为，跟之前的变化在于，更换LinearSVM(PSSM)后，该特征的feature importance为1st
    # 把学习率调高，更换特征后，较低的学习率可能达到的并不是全局最优
    params = {
        'boosting_type': boosting_type,
        'objective': "lambdarank",
        'metric': 'ndcg',
        'num_leaves':11,
        'max_depth':4,
        'learning_rate': 0.059,
        'feature_fraction': 1.0,
        'bagging_fraction': 0.9,
        'bagging_freq': 1,
        "max_bin":127, # 减小以防止过拟合,但太小模型的性能也会不好
        "min_data_in_leaf":50,
        "min_sum_hessian_in_leaf":5.0,
        "is_enable_sparse":"true",
        "force_row_wise":"true",
        "lambda_l1":5.7,
        "lambda_l2":32.,
        "ndcg_eval_at":"50"
    }

    train_args = dict(
        num_boost_round = 173,
        verbose_eval = 5,
    )

    if not (boosting_type == "dart"):
        if use_test_flag or use_eval_flag:
            train_args["early_stopping_rounds"] = 100
    
    conf = dict(
        params = params,
        train_args = train_args,
        norm_flag = norm_flag,
        eval_norm = eval_norm, # 归一化对较难的数据集会有1%的提升，比如在验证集上
        use_eval_flag = use_eval_flag,
        cross_validation_path = "dataset/benchmark/cross/fold_num2sequence_name.csv",
        specific_path = "feature/hhblits/acc_hmm/sequence_name2row_num.csv",
        vspecific_path = "feature/hhblits/acc_hmm/val_sequence_name2row_num.csv",
        key2value_path = "dataset/SCOP/csv/sequence_superfamily.csv",
        representation_dirs = [
            "feature/multiclassification/training/hhblits/representation",
            "feature/multiclassification/training/psiblastexb/representation",
            "feature/S2C/training_cluster/hhblitsexb",
            "feature/S2C/training_cluster/hhsearchexb",
            "feature/S2C/training_cluster/psiblastexb",
            "feature/S2C/training_cluster/deltablast"
        ],
        save_booster_l = save_booster_l
    )
    return conf

# n 表示n折交叉验证
def training_config(n, **kArgs):
    cross_validation_path = kArgs["cross_validation_path"]
    specific_path = kArgs["specific_path"]
    vspecific_path = kArgs["vspecific_path"]
    key2value_path = kArgs["key2value_path"]
    representation_dirs = kArgs["representation_dirs"]

    cv_df = pd.read_csv(cross_validation_path)
    spec_df = pd.read_csv(specific_path,index_col=0)
    vspec_df = pd.read_csv(vspecific_path,index_col=0)
    kv_df = pd.read_csv(key2value_path)

    template_df = spec_df.reset_index().rename(columns={"index":"template"})
    vtemplate_df = vspec_df.reset_index().rename(columns={"index":"template"})

    representation_dict = {}
    for i in range(n):
        representation_dict[i] = [
            os.path.join(x,"cv0{}.npy".format(i)) \
                for x in representation_dirs
        ]
    vrepresentation_dict = {}
    for i in range(n):
        vrepresentation_dict[i] = [
            os.path.join(x,"val0{}.npy".format(i)) \
                for x in representation_dirs
        ]
    
    k_df = template_df.join(kv_df.set_index("template"),on="template").loc[:,["template","sf","label"]]
    
    kArgs["template_df"] = template_df
    kArgs["vtemplate_df"] = vtemplate_df
    kArgs["cv_df"] = cv_df
    kArgs["k_df"] = k_df
    kArgs["kv_df"] = kv_df
    kArgs["representation_dict"] = representation_dict
    kArgs["vrepresentation_dict"] = vrepresentation_dict

    return kArgs

# 一般的训练过程不会对con, label进行额外的处理
@gen_training_ltr_template
def gen_training_ltr(con, label, **kArgs):
    return con, label

@ training_ltr_template
def training_ltr(i, **kArgs):
    return gen_training_ltr(i, **kArgs)

def test_config(n,eval_norm, save_output_flag,save_flag,save_booster_l,eval_print_path):
    cross_validation_path = "dataset/benchmark/cross/fold_num2sequence_name.csv"
    specific_path = "feature/hhblits/acc_hmm/sequence_name2row_num.csv"
    key2value_path = "dataset/SCOP/csv/sequence_superfamily.csv"
    representation_dirs = [
        "feature/multiclassification/hhblits/representation",
        "feature/multiclassification/psiblastexb/representation",
        "feature/S2C/test/hhblitsexb",
        "feature/S2C/test/hhblitsexb",
        "feature/S2C/test/psiblastexb",
        "feature/S2C/test/deltablast"
    ]
    spec_df = pd.read_csv(specific_path,index_col=0)
    kv_df = pd.read_csv(key2value_path)
    template_df = spec_df.reset_index().rename(columns={"index":"template"})

    conf = dict(
        template_df = template_df,
        cv_df = pd.read_csv(cross_validation_path),
        k_df = template_df.join(kv_df.set_index("template"),on="template").loc[:,["template","sf","label"]],
        kv_df = kv_df,
        save_booster_l = save_booster_l,
        eval_norm = eval_norm,
        save_output_flag = save_output_flag,
        eval_print_path = eval_print_path,
        save_flag = save_flag
    )

    representation_dict = {}
    for i in range(n):
        representation_dict[i] = [
            os.path.join(x,"cv0{}.npy".format(i)) \
                for x in representation_dirs
        ]
    
    save_l = [
        "feature/LTR/representation/cv0{}.npy".format(x) \
            for x in range(n)
    ]
    output_file_path = "evaluation/roc_at/lightgbm/lambdarank/representation/eval_seq_rocat.txt"

    conf["representation_dict"] = representation_dict
    conf["save_l"] = save_l
    conf["output_file_path"] = output_file_path
    return conf

# gen_clb是一个函数类型，函数签名为(i, **kArgs) -> (Tenser, Tenser, Booster)
@gen_test_ltr_template
def gen_test_clb(i, **kArgs):
    fn_col = str(i)
    cv_df = kArgs["cv_df"]
    template_df = kArgs["template_df"]
    representation_dict = kArgs["representation_dict"]
    kv_df = kArgs["kv_df"]
    kv_i_df = kArgs["kv_i_df"]
    representation_l = representation_dict[i]
    query_df = template_df.loc[template_df.loc[:,"template"].isin(cv_df.loc[:,fn_col])]

    con = run_lambdamart.get_training_set(representation_l)
    label = run_lambdamart.get_label(kv_df, kv_i_df, query_df,keepdims=True)
    return con, label

@test_ltr_template
def test_ltr(i,**kArgs):
    return gen_test_clb(i, **kArgs)

def val_config(n,eval_norm, save_output_flag,save_flag,save_booster_l,eval_print_path):
    cross_validation_path = "dataset/benchmark/cross/fold_num2sequence_name.csv"
    specific_path = "feature/hhblits/acc_hmm/sequence_name2row_num.csv"
    vspecific_path = "feature/hhblits/acc_hmm/val_sequence_name2row_num.csv"
    key2value_path = "dataset/SCOP/csv/sequence_superfamily.csv"
    representation_dirs = [
        "feature/multiclassification/training/hhblits/representation",
        "feature/multiclassification/training/psiblastexb/representation",
        "feature/S2C/training_cluster/hhblitsexb",
        "feature/S2C/training_cluster/hhsearchexb",
        "feature/S2C/training_cluster/psiblastexb",
        "feature/S2C/training_cluster/deltablast"
    ]
    spec_df = pd.read_csv(specific_path,index_col=0)
    vspec_df = pd.read_csv(vspecific_path,index_col=0)
    kv_df = pd.read_csv(key2value_path)
    template_df = spec_df.reset_index().rename(columns={"index":"template"})
    vtemplate_df = vspec_df.reset_index().rename(columns={"index":"template"})

    conf = dict(
        vtemplate_df = vtemplate_df,
        cv_df = pd.read_csv(cross_validation_path),
        k_df = template_df.join(kv_df.set_index("template"),on="template").loc[:,["template","sf","label"]],
        kv_df = kv_df,
        save_booster_l = save_booster_l,
        eval_norm = eval_norm,
        save_output_flag = save_output_flag,
        eval_print_path = eval_print_path,
        save_flag = save_flag
    )

    representation_dict = {}
    for i in range(n):
        representation_dict[i] = [
            os.path.join(x,"val0{}.npy".format(i)) \
                for x in representation_dirs
        ]
    
    save_l = [
        "feature/LTR/training/representation/val0{}.npy".format(x) \
            for x in range(n)
    ]
    output_file_path = "evaluation/roc_at/lightgbm/lambdarank/representation/val_seq_rocat.txt"

    conf["representation_dict"] = representation_dict
    conf["save_l"] = save_l
    conf["output_file_path"] = output_file_path
    return conf

# 这里的key2value和其他地方用得不一样
# 因为要兼顾scop和scope数据集
def ind_config(n,eval_norm, save_output_flag,save_flag,save_booster_l,eval_print_path):
    cross_validation_path = "dataset/benchmark/cross/fold_num2sequence_name.csv"
    key2value_path = "dataset/scop_scope/csv/sequence_superfamily.csv"
    specific_path = "feature/hhblits/acc_hmm/sequence_name2row_num.csv"
    ispecific_path = "feature/hhblits/acc_hmm/ind_sequence_name2row_num.csv"
    output_file_path = "evaluation/roc_at/lightgbm/lambdarank/representation/ind_seq_rocat.txt"
    representation_dirs = [
        "feature/multiclassification/training/hhblits/representation",
        "feature/multiclassification/training/psiblastexb/representation",
        "feature/S2C/training_cluster/hhblitsexb",
        "feature/S2C/training_cluster/hhsearchexb",
        "feature/S2C/training_cluster/psiblastexb",
        "feature/S2C/training_cluster/deltablast"
    ]
    representation_dict = {}
    for i in range(n):
        representation_dict[i] = [
            os.path.join(x,"ind0{}.npy".format(i)) \
                for x in representation_dirs
        ]
    spec_df = pd.read_csv(specific_path,index_col=0)
    ispec_df = pd.read_csv(ispecific_path,index_col=0)
    template_df = spec_df.reset_index().rename(columns={"index":"template"}) 
    kv_df = pd.read_csv(key2value_path)   

    conf = dict(
        cv_df = pd.read_csv(cross_validation_path),
        k_df = template_df.join(kv_df.set_index("template"),on="template").loc[:,["template","sf","label"]],
        kv_df = kv_df,
        itemplate_df = ispec_df.reset_index().rename(columns={"index":"template"}),
        representation_dict = representation_dict,
        eval_norm = eval_norm,
        save_output_flag = save_output_flag,
        save_flag = save_flag,
        save_booster_l = save_booster_l,
        output_file_path = output_file_path,
        eval_print_path = eval_print_path,
        save_l = [
            "feature/LTR/training/representation/ind0{}.npy".
        format(x) \
                for x in range(n)
        ]
    )
    return conf

@gen_test_ltr_template
def gen_val_clb(i, **kArgs):
    kv_df = kArgs["kv_df"]
    kv_i_df = kArgs["kv_i_df"]
    vtemplate_df = kArgs["vtemplate_df"]
    representation_dict = kArgs["representation_dict"]
    representation_l = representation_dict[i]

    con = run_lambdamart.get_training_set(representation_l)
    label = run_lambdamart.get_label(kv_df,kv_i_df, vtemplate_df,keepdims=True)

    return con, label

@test_ltr_template
def val_ltr(i, **kArgs):
    return gen_val_clb(i, **kArgs)

@gen_test_ltr_template
def gen_ind_clb(i, **kArgs):
    representation_dict = kArgs["representation_dict"]
    representation_l = representation_dict[i]
    kv_df = kArgs["kv_df"]
    kv_i_df = kArgs["kv_i_df"]
    itemplate_df = kArgs["itemplate_df"]
    con = run_lambdamart.get_training_set(representation_l)
    label = run_lambdamart.get_label(kv_df,kv_i_df, itemplate_df,keepdims=True)

    return con, label

@test_ltr_template
def ind_ltr(i,**kArgs):
    return gen_ind_clb(i, **kArgs)

# kArgs要指定当前对应的sequence identity
# 存放在kArgs["diff"]中
# 注意kArgs["diff_path"]保存的是csv格式的不同序列对应的子集编号
# 当前sequence identity对应数据集中，根据交叉验证划分结果，
# 用于训练的序列为除去属于测试集的部分
# 相当于在原本的训练集中选择属于对应数据集中的序列
@gen_training_ltr_template
def gen_straining_ltr(con,label,**kArgs):
    t_i_df = kArgs["t_i_df"]
    diff = kArgs["diff"]
    diff_path = kArgs["diff_path"]
    # 提取属于当前sequence identity下的序列
    # 根据label重新编号，得到new_row_num
    diffdf = pd.read_csv(diff_path)
    t_i_df = t_i_df.assign(new_row_num=list(range(t_i_df.shape[0])))
    sel_i_df = t_i_df.loc[t_i_df.loc[:,"template"].isin(diffdf.loc[diffdf.loc[:,diff],"template"])]
    sel_row_num = sel_i_df.loc[:,"new_row_num"]
    indices = tf.constant(sel_row_num.to_list(),dtype=tf.int32)
    indices = tf.reshape(indices,(-1,1))
    con = tf.gather_nd(con,indices)
    label = tf.gather_nd(label,indices)

    return (con, label)

@training_ltr_template
def straining_ltr(i, **kArgs):
    return gen_straining_ltr(i, **kArgs)

@gen_test_ltr_template
def gen_sind_clb(i, **kArgs):
    representation_dict = kArgs["representation_dict"]
    representation_l = representation_dict[i]
    kv_df = kArgs["kv_df"]
    kv_i_df = kArgs["kv_i_df"]
    itemplate_df = kArgs["itemplate_df"]
    diff = kArgs["diff"]
    diff_path = kArgs["diff_path"]
    # 提取属于当前sequence identity下的序列
    # 根据label重新编号，得到new_row_num
    diffdf = pd.read_csv(diff_path)

    con = run_lambdamart.get_training_set(representation_l)
    label = run_lambdamart.get_label(kv_df,kv_i_df, itemplate_df,keepdims=True)

    itemplate_df = itemplate_df.assign(new_row_num=list(range(itemplate_df.shape[0])))
    sel_i_df = itemplate_df.loc[itemplate_df.loc[:,"template"].isin(diffdf.loc[diffdf.loc[:,diff],"template"])]
    sel_row_num = sel_i_df.loc[:,"new_row_num"]
    indices = tf.constant(sel_row_num.to_list(),dtype=tf.int32)
    indices = tf.reshape(indices,(-1,1))
    con = tf.gather_nd(con,indices)
    label = tf.gather_nd(label,indices)

    return con, label

@test_ltr_template
def sind_ltr(i, **kArgs):
    return gen_sind_clb(i, **kArgs)

    
# 这里要把diff_path对应的csv文件读取出
# 建立dataframe，并把所有sequence identity通过list格式保存在kArgs["all_diff"]中
# dataframe包含这几列，[1:]为所有sequence identity
# Index(['template', '0.4', '0.45', '0.5', '0.55', '0.6', '0.65', '0.7', '0.75',
#       '0.8', '0.85', '0.9']
# 新增0.25, 0.3, 0,35这三列
def seqiden_all_config(*args):
    n,boosting_type,use_eval_flag,\
    use_test_flag,norm_flag,eval_norm,\
    save_output_flag,save_flag,\
    save_booster_l,\
    sequence_identity_path,eval_print_path = args

    config = init_training_config(boosting_type,use_eval_flag,use_test_flag,
      norm_flag,eval_norm,save_booster_l)
    config = training_config(n,**config)
    config["diff_path"] = sequence_identity_path
    diff_path = sequence_identity_path
    diffdf = pd.read_csv(diff_path)
    config["all_diff"] = diffdf.columns[1:]
    # 补充保存feature importance的位置
    save_feature_importance_l = [
        "evaluation/feature_importance/sequence_identity/seq_iden_{}.txt".format(d) \
            for d in diffdf.columns[1:]
    ]
    config["save_feature_importance_l"] = save_feature_importance_l

    iconfig = ind_config(n,eval_norm,save_output_flag,
        save_flag,config["save_booster_l"],eval_print_path)
    iconfig["diff_path"] = sequence_identity_path
    iconfig["save_feature_importance_l"] = save_feature_importance_l
    iconfig["all_diff"] = diffdf.columns[1:]


    return config, iconfig

# 这里每次要更新模型保存的名称
# 先实现训练部分
def seqident_training(n, **kArgs):
    all_diff = kArgs["all_diff"]

    def run_diff(diff, **kArgs):
        kArgs["diff"] = diff
        dir = os.path.dirname(kArgs["save_booster_l"][0])
        kArgs["save_booster_l"] = [
            os.path.join(dir,"seq_iden_{}_0{}.txt".format(diff,x)) \
                for x in range(n)
        ]
        straining_ltr(n,**kArgs)

    list(map(lambda d: run_diff(d,**kArgs), all_diff))

def seqident_ind(n, **kArgs):
    all_diff = kArgs["all_diff"]
    dir = os.path.dirname(kArgs["save_booster_l"][0])

    def test_diff(diff):
        kArgs["diff"] = diff
        kArgs["save_booster_l"] = [
            os.path.join(dir,"seq_iden_{}_0{}.txt".format(diff,x)) \
                for x in range(n)
        ]
        ind_ltr(n,**kArgs)
    
    list(map(test_diff,all_diff))

def sind_feature_importance(n, **kArgs):
    all_diff = kArgs["all_diff"]
    dir = os.path.dirname(kArgs["save_booster_l"][0])
    save_feature_importance_l = kArgs["save_feature_importance_l"]

    def get_feature_importance(booster_path):
        booster = lgb.Booster(model_file=booster_path)
        fi = booster.feature_importance()
        return fi

    def save_diff_feature_importance(*args):
        diff, feature_importance_path = args
        save_booster_l = [os.path.join(dir,"seq_iden_{}_0{}.txt".format(diff,x)) for x in range(n)]
        fary = np.array(list(map(get_feature_importance,save_booster_l)))
        np.savetxt(feature_importance_path,fary,"%.2f")

    list(map(lambda p: save_diff_feature_importance(*p),zip(all_diff,save_feature_importance_l)))

# 以上为实验内容
#-------------------------------------------------------------------------------------------------------------------------------
#%%
if __name__ == "__main__":
    all_fold = 5 # 5折交叉验证
    the_boosting_type = "gbdt"
    # the_boosting_type = "dart"
    auto_timestamp = True # 自动生成一个timestampe区分不同的模型
    need_use_eval_flag = True # 用验证集找最优tree, 结果测试的时候可以用False
    need_use_test_flag = False # 用测试集做验证集
    need_norm_flag = False # 训练集归一化，没必要
    need_eval_norm = False # 归一化对较难的数据集会有1%的提升，比如在验证集上
    need_save_output_flag = False # 保存每个序列的评估结果
    need_save_flag = False # 保存LTR的排序结果, 如果不需要的话可以选为False
    need_test_diff = False # 测试在不同sequence identity下, feature importance与roc1, roc50的关系

    if auto_timestamp:
        time_suffix = time.strftime("%Y%m%d-%H:%M:%S",time.localtime())
    else:
        time_suffix = "20201022-07:44:12" # 固定timestampe用于测试某个时间点得到的模型

    if the_boosting_type == "dart":
        dir_name = "dart"
    else:
        dir_name = "lambdarank"
    
    if need_test_diff:
        sel_name = "diff"
    else:
        # 这两个目录都是可以用的
        # mix中最晚的一种方法结果最好
        # new中结果只比mix中好一点，这也验证了，单一特征的性能并不能影响整体
        # sel_name = "mix"
        sel_name = "new"

    model_save_path = [
            "model/{}/lightgbm/{}/representation/cv0{}-{}.txt".format(sel_name, dir_name,x,time_suffix)\
                for x in range(all_fold)
    ]

    the_eval_print_path = "output/lgb/lgb_diff_eval_{}".format(time_suffix)
    scop_fas_path = "dataset/benchmark/all/SCOP_Experimental"
    # 高于0.4的情况
    the_sequence_identity_path = "dataset/scop_scope/csv/sequence_identity.csv"
    the_single_seqiden_path = "dataset/scop_scope/csv/single_sequence_identity.csv"
    # 低于0.4的情况
    the_seqiden_low_path = "dataset/scop_scope/csv/sequence_identity_low.csv"
    the_sgle_seqiden_low_path = "dataset/scop_scope/csv/single_sequence_identity_low.csv"


    # 准备----------------------------------------------------------------
    # config = test_sequence_identity()
    # sequence2benchmark(the_seqiden_low_path,**config)
    # # 单个情况
    # config = single_sequence_identity(the_sgle_seqiden_low_path,the_seqiden_low_path)
    # call_single_seq2bench(**config)
    # call_add_sequence_identity(**config)
    # 合并两种情况
    df_high = pd.read_csv(the_sequence_identity_path)
    df_low = pd.read_csv(the_seqiden_low_path)
    df_high.loc[:,df_low.columns[1:]] = df_low.loc[:, df_low.columns[1:]]
    df_high = pd.concat(
        [
            df_high.iloc[:,[0]],
            df_high.loc[:,df_high.columns[1:].sort_values()]
        ],
        axis=1
    )
    df_high.to_csv(the_sequence_identity_path,index=False)
    # --------------------------------------------------------------------

    # LTR----------------------------------------------------------------
    # 需要使用这部分代码时取消注释即可

    # # 训练
    # config = init_training_config(the_boosting_type,need_use_eval_flag,need_use_test_flag,
    #   need_norm_flag,need_eval_norm,model_save_path)
    # config = training_config(all_fold,**config)
    # training_ltr(all_fold,**config)
    # # 测试
    # config = test_config(all_fold,need_eval_norm,need_save_output_flag,need_save_flag,model_save_path,the_eval_print_path)
    # test_ltr(all_fold,**config)
    # # 验证集
    # config = val_config(all_fold,need_eval_norm,need_save_output_flag,need_save_flag,model_save_path,the_eval_print_path)
    # val_ltr(all_fold,**config)
    # # 独立测试集
    # config = ind_config(all_fold,need_eval_norm,need_save_output_flag,need_save_flag,model_save_path,the_eval_print_path)
    # ind_ltr(all_fold,**config)

    #---------------------------------------------------------------------
    # test sequence identity----------------------------------------------
    # 测试在不同sequence identity对应的数据集下，feature importances与roc1, roc50的关系
    # 在独立测试集上进行测试
    
    # diff_config, dd_config = seqiden_all_config(all_fold,the_boosting_type,need_use_eval_flag,
    #     need_use_test_flag,need_norm_flag,need_eval_norm,
    #     need_save_output_flag,need_save_flag,
    #     model_save_path,the_sequence_identity_path,the_eval_print_path)
    
    # seqident_training(all_fold,**diff_config)
    # seqident_ind(all_fold,**dd_config)
    # sind_feature_importance(all_fold,**dd_config)

    # diff_config里面一定要包含新序列对应row num的dataframe，因为，原来生成用于训练的矩阵就是符合row num顺序的
    # 而里面也存在序列对应row num的dataframe，这样就可以取出需要的序列用于训练
    #---------------------------------------------------------------------
    # 只测试一个

    # diff_config, dd_config = seqiden_all_config(all_fold,the_boosting_type,need_use_eval_flag,
    #     need_use_test_flag,need_norm_flag,need_eval_norm,
    #     need_save_output_flag,need_save_flag,
    #     model_save_path,the_single_seqiden_path,the_eval_print_path)
    
    # seqident_training(all_fold,**diff_config)
    # seqident_ind(all_fold,**dd_config)
    # sind_feature_importance(all_fold,**dd_config)

    #---------------------------------------------------------------------

#%%
import pandas as pd

    # 高于0.4的情况
the_sequence_identity_path = "dataset/scop_scope/csv/sequence_identity.csv"
the_single_seqiden_path = "dataset/scop_scope/csv/single_sequence_identity.csv"
# 低于0.4的情况
the_seqiden_low_path = "dataset/scop_scope/csv/sequence_identity_low.csv"
the_sgle_seqiden_low_path = "dataset/scop_scope/csv/single_sequence_identity_low.csv"


# 准备----------------------------------------------------------------
# config = test_sequence_identity()
# sequence2benchmark(the_seqiden_low_path,**config)
# # 单个情况
# config = single_sequence_identity(the_sgle_seqiden_low_paththe_seqiden_low_path)
# call_single_seq2bench(**config)
# call_add_sequence_identity(**config)
# 合并两种情况
df_high = pd.read_csv(the_sequence_identity_path)
df_low = pd.read_csv(the_seqiden_low_path)
df_high.loc[:,df_low.columns[1:]] = df_low.loc[:, df_low.columns[1:]]
df_high = pd.concat([df_high.iloc[:,[0]],df_high.loc[:,df_high.columns[1:].sort_values()]],axis=1)
df_high
# %%
