# To add a new cell, type '# %%'
# To add a new markdown cell, type '# %% [markdown]'
# %% [markdown]
# ## 尝试进行特征提取
# ### 1. 尝试psiblast
# > 使用psiblast时，尝试用不同的参数，观察输出结果

# %%
from src.utils.feature_extraction import PsiBlast

run_psiblast = PsiBlast()

params = {
    "db": "dataset/benchmark/db/template/scop_0_psiblast",
    "query": "dataset/benchmark/seqs/SCOP_Experimental_foldNo0/d1a8qa_.seq.ffindex",
    "out": "dataset/benchmark/tmp/d1a8qa_fmt7_add.psi",
    "evalue": 1e-4,
    "outfmt": "'7 qaccver saccver pident evalue bitscore'",
    "np": 10
}

run_psiblast.use_psiblast(params)

# %% [markdown]
# > 测试没有hit出现的情况

# %%
import subprocess as sp

from src.utils.feature_extraction import PsiBlast

run_psiblast = PsiBlast()

tmpFastaFilePath = "dataset/benchmark/tmp/fasta合集"
tmpBlastdbPath = "dataset/benchmark/tmp/fasta"
makeblastdb_cmd = "makeblastdb -in {} -input_type fasta -dbtype prot -out {}".format(
    tmpFastaFilePath,tmpBlastdbPath
)

r = sp.call(makeblastdb_cmd,shell=True)
if r != 0:
    print("error code is {}".format(r))

params = {
    "db": tmpBlastdbPath,
    "query": "dataset/benchmark/seqs/SCOP_Experimental_foldNo0/d1a6fa_.seq",
    "out": "dataset/benchmark/tmp/d1a6fa_empty.psi",
    "evalue": 1e-4,
    "outfmt": "'7 qaccver saccver pident evalue bitscore'",
    "num_iterations": 3
}

run_psiblast.use_psiblast(params)

# %% [markdown]
# > 尝试解析，如果这里能够成功，就把这里的结果放到feature_extraction中

# %%
import re

# psi_path = "dataset/benchmark/tmp/d1a6fa_fmt7_add.psi"
psi_path = "dataset/benchmark/tmp/d1an7a_fmt7_add.psi"

start = re.compile(r"# [0-9]+ (hit|hits) found")
contentStart = re.compile(r"^[^#\s]")
end = re.compile(r"# BLAST processed [0-9]+ (query|queries)")


getContentFlag = False
with open(psi_path,"r") as h:
    for line in h:
        if not getContentFlag:
            start_r = start.search(line)
            if start_r is not None:
                content_l = []
                getContentFlag = True
        else:
            end_r = end.search(line)
            if end_r is not None:
                break
            cr = contentStart.search(line)
            if cr is not None:
                content_l.append(line)
            else:
                getContentFlag = False

contentParser = re.compile(r"\t")
parsing_dict = {"query":[],"template":[],
    "percentage of identical matches":[],"bit score":[]}
for c in content_l:
    parsing_result = contentParser.split(c)
    parsing_dict["query"].append(parsing_result[0])
    parsing_dict["template"].append(parsing_result[1])
    parsing_dict["percentage of identical matches"].append(parsing_result[2])
    parsing_dict["bit score"].append(parsing_result[3])
parsing_dict


# %%
from src.utils.feature_extraction import PsiBlast

run_psiblast = PsiBlast()

parsing_dict = {"query":[],"template":[],
    "percentage of identical matches":[],
    "expect value":[],"bit score":[]}

psi_path = "dataset/benchmark/tmp/d1an7a_fmt7_add.psi"
parsing_dict = run_psiblast.parsing_searched_result(psi_path)

# psi_path = "dataset/benchmark/tmp/d1an7a_fmt7_add.psi"
# psi_path = "dataset/benchmark/tmp/d1a6fa_empty.psi"
psi_path = "dataset/benchmark/tmp/d1a6fa_empty.psi"
parsing_dict = run_psiblast.parsing_searched_result(psi_path,parsing_dict)

print(parsing_dict)

# %% [markdown]
# > 通过psiblast对template查询，一边查询就一边进行解析，解析结果存放为csv文件

# %%
from src.utils.feature_extraction import PsiBlast
# import pandas as pd
from mars import dataframe as md
import os

run_psiblast = PsiBlast()

n = 5
fasta_l = ["dataset/benchmark/seqs/SCOP_Experimental_foldNo{}".format(x)     for x in range(n)]
db_l = ["dataset/benchmark/db/template/scop_{}_psiblast".format(x)     for x in range(n)]
outPath = "feature/tmp/searched.psi"

parsing_dict = {"query":[],"template":[],
    "percentage of identical matches":[],
    "expect value":[],"bit score":[]}
limit = 5 #用于简单测试
testFlag = False
mask_l = [False,False,False,False,True]
for i, fastaDir in enumerate(fasta_l):
    if not mask_l[i]:
        continue
    dbPath = db_l[i]
    for fastaName in os.listdir(fastaDir):
        if testFlag and limit == 0:
            break
        fastaFilePath = os.path.join(fastaDir,fastaName)
        params = {
            "db": dbPath,
            "query": fastaFilePath,
            "out": outPath,
            "evalue": 1e-4,
            "outfmt": "'7 qaccver saccver pident evalue bitscore'",
            "num_iterations": 3
        }
        run_psiblast.use_psiblast(params)
        parsing_dict = run_psiblast.parsing_searched_result(outPath,parsing_dict)
        if testFlag:
            limit -= 1

# parsing_df = pd.DataFrame(parsing_dict)
# parsing_df
parsing_df = md.DataFrame(parsing_dict)
# parsing_df.execute()

# %% [markdown]
# > 保存成csv文件

# %%
csvPath = "feature/psiblast/csv/searched_fold04.csv"

r = parsing_df.to_csv(csvPath,index=False).execute()
print(r)

# %% [markdown]
# > 检验空值是否成功插入

# %%
parsing_df = md.read_csv(csvPath)
parsing_df.execute()

# %% [markdown]
# > 把searched.csv核searched_fold04.csv合并起来

# %%
csvPath = "feature/psiblast/csv/searched.csv"
parsing_df = md.read_csv(csvPath)
csvPath = "feature/psiblast/csv/searched_fold04.csv"
pFold_4_df = md.read_csv(csvPath)

print(parsing_df.shape)
print(pFold_4_df.shape)


# %%


# %% [markdown]
# > 根据5折交叉验证所涉及的所有pairwise，制作一个表格，格式为query, template。这样方便后续对不同方法进行性能评估

# %%
import pandas as pd
from Bio.SeqIO import parse
from Bio.Alphabet import generic_protein
import re

def combinations(L, k):
    if k == 0:
        return [[]]
    if L == []:
        return []
    
    x = L[0]
    res = L[1:]

    part = [[x] + y for y in combinations(res,k-1)]
    return part + combinations(res,k)

combs = combinations(list(range(5)),4)
combs.reverse()

n = 5
fasta_l = ["dataset/benchmark/cross/SCOP_Experimental_foldNo{}".format(x)     for x in range(n)]

pairwise_dict = {"query":[],"template":[]}

for i, c in enumerate(combs):
    queryFastaPath = fasta_l[i]
    with open(queryFastaPath, "r") as h:
        for qRecord in parse(h,"fasta",generic_protein):
            queryName = qRecord.name
            for j in c:
                templateFastaPath = fasta_l[j]
                with open(templateFastaPath,"r") as h1:
                    for tRecord in parse(h1,"fasta",generic_protein):
                        templateName = tRecord.name
                        pairwise_dict["query"].append(queryName)
                        pairwise_dict["template"].append(templateName)

pairwise_df = pd.DataFrame(pairwise_dict)
        

# %% [markdown]
# > 保存该结果

# %%
csvPath = "feature/common.csv"
with open(csvPath,"w",newline='') as h:
    pairwise_df.to_csv(h,index=False)

# %% [markdown]
# > 评估psiblast的检索结果，roc1，先在这里测试之后再放到feature_extraction中
# >> 计算roc1的方法实际上是选取第1个false positive前的查询结果，并绘制roc曲线，计算面积

# %%
from sklearn.metrics import roc_curve

fpr, tpr = roc_curve()

# %% [markdown]
# > 筛选部分序列，这些序列是第一个false positive前的序列

# %%
from six.moves import cPickle

seq2sf_path = "dataset/benchmark/dict/SCOP_Experimental/sf/binary/key2value"
with open(seq2sf_path,"rb") as h:
    seq2sf = cPickle.load(h)

def is_same_sf(seq1,seq2):
    return seq2sf[seq1] == seq2sf[seq2]

index_l = []
skip_flag = False
isSameSf = True
query_set = set()
for i in parsing_df.index:
    query_set.add(parsing_df.loc[i,"query"])
    isSameSf = is_same_sf(parsing_df.loc[i,"query"],parsing_df.loc[i,"template"])
    if skip_flag:
        if parsing_df.loc[i,"query"] not in query_set:
            skip_flag = False
        else:
            continue
    
    if not skip_flag and isSameSf:
        index_l.append(i)
    
    if not skip_flag and not isSameSf:
        skip_flag = True
        continue
    


filtered_df = parsing_df.loc[index_l]
filtered_df

# %% [markdown]
# ### 尝试用mmseq2

# %%
import os
import subprocess as sp

seqPath = "/home/shaojiangyi/source/Representative_Rank/dataset/benchmark/seqs/SCOP_Experimental_foldNo1/d1a25a_.seq"
dbPath = "dataset/benchmark/db/template/scop_1_mmseqs"
mmrPath = "dataset/benchmark/tmp/d1a25a_.m8"
tmpDir = "/home/shaojiangyi/source/Representative_Rank/dataset/benchmark/tmp/tmp"

if os.path.exists(tmpDir):
    cmd = "rm -r {}".format(tmpDir)
    sp.call(cmd,shell=True)
else:
    os.mkdir(tmpDir)

cmd = "mpirun -np 10 mmseqs easy-search {} {} {} {} -e 1e-4 --remove-tmp-files 1".format(seqPath,dbPath,mmrPath,tmpDir)
r = sp.call(cmd,shell=True)

if os.path.exists(tmpDir):
    cmd = "rm -r {}".format(tmpDir)
    sp.call(cmd,shell=True)

if r != 0:
    print("error code {}".format(r))

# %% [markdown]
# > 使用mmseqs的search

# %%
seqPath = "dataset/benchmark/seqs/SCOP_Experimental_foldNo1/d1a25a_.seq"
sDbPath = "dataset/benchmark/tmp/seqDB"
dbPath = "dataset/benchmark/db/template/scop_1_mmseqs"
mmrPath = "dataset/benchmark/tmp/d1a25a_-v1.mmr"
tmpDir = "/home/shaojiangyi/source/Representative_Rank/dataset/benchmark/tmp/tmp"

if os.path.exists(tmpDir):
    cmd = "rm -r {}".format(tmpDir)
    sp.call(cmd,shell=True)
else:
    os.mkdir(tmpDir)

cmd = "mmseqs createdb {} {}".format(seqPath,sDbPath)
sp.call(cmd,shell=True)

cmd = "mmseqs search {} {} {} {}".format(sDbPath, dbPath, mmrPath,tmpDir)
sp.call(cmd,shell=True)

if os.path.exists(tmpDir):
    cmd = "rm -r {}".format(tmpDir)
    sp.call(cmd,shell=True)

# %% [markdown]
# > 在尝试下用feature_extraction里的mmseqs easy-search

# %%
import os
import subprocess as sp
from src.utils.feature_extraction import MMseqs2

runMMseqs = MMseqs2()

seqPath = "dataset/benchmark/seqs/SCOP_Experimental_foldNo1/d1wi6a1.seq"
dbPath = "dataset/benchmark/db/template/scop_1_mmseqs"
m8Path = "dataset/benchmark/tmp/d1wi6a1.m8"
tmpDir = "/home/shaojiangyi/source/Representative_Rank/dataset/benchmark/tmp/tmp"

if os.path.exists(tmpDir):
    cmd = "rm -r {}".format(tmpDir)
    sp.call(cmd,shell=True)
else:
    os.mkdir(tmpDir)

set_params = {
    "queryDBPath":seqPath,
    "targetDBPath": dbPath,
    "alignFilePath": m8Path,
    "tmpPath": tmpDir,
    "evalue": 1e-4,
    "mpi_runner": 10
}

runMMseqs.easy_search(set_params)

if os.path.exists(tmpDir):
    cmd = "rm -r {}".format(tmpDir)
    sp.call(cmd,shell=True)

# %% [markdown]
# > 但是mmseqs search的结果没有mmseqs easy-search的方便访问与解析, 接下来尝试解析mmseqs的排序结果

# %%
import re
import pandas as pd

m8Path_l = ["dataset/benchmark/tmp/d1a25a_.m8",
    "dataset/benchmark/tmp/d1nkua_.m8",
    "dataset/benchmark/tmp/d1wi6a1.m8"]

parsing_dict = {"query":[],"template":[],
    "prob":[],"evalue":[],"score":[]}

# d1a25a_	d1dsya_	0.765	132	31	0	1	132	2	133	9.023E-68	221
for m8Path in m8Path_l:
    spliter = re.compile(r"\s+")
    with open(m8Path,"r") as h:
        for line in h:
            L = spliter.split(line)
            query, template,prob = L[:3]
            evalue, score = L[-2:]
            parsing_dict["query"].append(query)
            parsing_dict["template"].append(template)
            parsing_dict["prob"].append(prob)
            parsing_dict["evalue"].append(evalue)
            parsing_dict["score"].append(score)
        if parsing_dict["query"] == []:
            fileName_pr = re.compile(r"\.")
            fileName_r = fileName_pr.search(os.path.basename(m8Path))
            if fileName_r is None:
                query = os.path.basename(m8Path)[:fileName_r.start()]
                parsing_dict["query"].append(query)
                parsing_dict["template"].append(pd.NA)
                parsing_dict["prob"].append(pd.NA)
                parsing_dict["evalue"].append(pd.NA)
                parsing_dict["score"].append(pd.NA)
        


